import json
from array import array
from itertools import chain, combinations, islice, product, tee
from typing import Callable, Dict, FrozenSet, Iterable, Iterator, List, Optional, Tuple

from comb_spec_searcher import CombinatorialObject
from permuta import Perm
from permuta.misc import DIR_EAST, DIR_NORTH, DIR_SOUTH, DIR_WEST, HTMLViewer, UnionFind

Cell = Tuple[int, int]
Position = Tuple[Cell, ...]


class GriddedPerm(CombinatorialObject):
    def __init__(
        self, pattern: Iterable[int] = (), positions: Iterable[Cell] = ()
    ) -> None:
        self._patt = Perm(pattern)
        self._pos = tuple(positions)
        if len(self._patt) != len(self._pos):
            raise ValueError("Pattern and position list have unequal lengths.")
        self._cells: FrozenSet[Cell] = frozenset(self._pos)

    @classmethod
    def single_cell(cls, pattern: Iterable[int], cell: Cell) -> "GriddedPerm":
        """Construct a gridded permutation where the cells are all located in a
        single cell."""
        pattern = Perm(pattern)
        return cls(pattern, (cell for _ in range(len(pattern))))

    @classmethod
    def empty_perm(cls) -> "GriddedPerm":
        """Construct the empty gridded permutation."""
        return cls((), ())

    @classmethod
    def point_perm(cls, cell: Cell) -> "GriddedPerm":
        """Construct the point gridded permutation using the cell."""
        return cls((0,), (cell,))

    def contradictory(self) -> bool:
        """Checks if the points of the griddedperm contradict the permutation.

        Checks if for every 0 <= i < j < n
            patt[i] <= patt[j]
            if patt[i] < patt[j] then pos[i] <= pos[j]
            if patt[i] > patt[j] then pos[i] >= pos[j]
        """
        return any(
            (l_x > r_x or (l_v < r_v and l_y > r_y) or (l_v > r_v and l_y < r_y))
            for j, (r_v, (r_x, r_y)) in enumerate(self)
            for l_v, (l_x, l_y) in islice(self, j)
        )

    def occupies(self, cell: Cell) -> bool:
        """Checks if the gridded permutation has a point in the given cell."""
        return cell in self._cells

    def occurrences_in(self, other: "GriddedPerm") -> Iterator[Tuple[int, ...]]:
        """Returns all occurrences of self in other."""
        yield from self._patt.occurrences_in(other.patt, self.pos, other.pos)

    def occurs_in(self, other: "GriddedPerm") -> bool:
        """Checks if self occurs in other."""
        return any(self.occurrences_in(other))

    def avoids(self, *patts: "GriddedPerm") -> bool:
        """Return true if self avoids all of the patts."""
        return not self.contains(*patts)

    def contains(self, *patts: "GriddedPerm") -> bool:
        """Return true if self contains an occurrence of any of patts."""
        return any(any(True for _ in patt.occurrences_in(self)) for patt in patts)

    def remove_cells(self, cells: Iterable[Cell]) -> "GriddedPerm":
        """Remove any points in the cell given and return a new gridded
        permutation."""
        cells = set(cells)
        remaining = [i for i, (_, pos) in enumerate(self) if pos not in cells]
        return type(self)(
            Perm.to_standard(self._patt[i] for i in remaining),
            (self._pos[i] for i in remaining),
        )

    def points_in_cell(self, cell: Cell) -> Iterator[int]:
        """Yields the indices of the points in the cell given."""
        return (i for i, (_, pos) in enumerate(self) if pos == cell)

    def isolated_cells(self) -> Iterator[Cell]:
        """Yields the cells that contain only one point of the gridded
        permutation and are in their own row and column."""
        return (
            (x1, y1)
            for i, (_, (x1, y1)) in enumerate(self)
            if not any(
                x1 == x2 or y1 == y2 for j, (_, (x2, y2)) in enumerate(self) if i != j
            )
        )

    def is_isolated(self, indices: Iterable[int]) -> bool:
        """Checks if the cells at the indices do not share a row or column with
        any other cell in the gridded permutation."""
        indices = set(indices)
        return not any(
            x == self._pos[j][0] or y == self._pos[j][1]
            for j in indices
            for i, (_, (x, y)) in enumerate(self)
            if i not in indices
        )

    def forced_point_index(self, cell: Cell, direction: int) -> int:
        """Search in the cell given for the point with the strongest force with
        respect to the given force."""
        if self.occupies(cell):
            indices = self.points_in_cell(cell)
            if direction == DIR_EAST:
                return max(indices)
            if direction == DIR_NORTH:
                return max((self._patt[idx], idx) for idx in indices)[1]
            if direction == DIR_WEST:
                return min(indices)
            if direction == DIR_SOUTH:
                return min((self._patt[idx], idx) for idx in indices)[1]
            raise ValueError("You're lost, no valid direction")
        raise ValueError("The gridded perm does not occupy the cell")

    def forced_point_of_requirement(
        self, gps: Tuple["GriddedPerm", ...], indices: Tuple[int, ...], direction: int
    ) -> Optional[Tuple[int, int]]:
        """
        Return the pair (x, y) where x is the gridded perm in gps that is
        farthest in the given direction, and y is index of the forced point
        with respect to the gps and indices. If gps is avoided, then
        return None.
        """

        def directionmost(i1: int, i2: int) -> int:
            """return the directionmost between i1 and i2."""
            if direction == DIR_EAST:
                return max(i1, i2)
            if direction == DIR_NORTH:
                return max((self._patt[p], p) for p in (i1, i2))[1]
            if direction == DIR_WEST:
                return min(i1, i2)
            if direction == DIR_SOUTH:
                return min((self._patt[p], p) for p in (i1, i2))[1]
            raise ValueError("You're lost, no valid direction")

        res: Optional[Tuple[int, int]] = None
        for idx, gp in enumerate(gps):
            forced_index_in_patt = indices[idx]
            for occurrence in gp.occurrences_in(self):
                if res is None:
                    res = idx, occurrence[forced_index_in_patt]
                else:
                    new_res = directionmost(res[1], occurrence[forced_index_in_patt])
                    if res[1] != new_res:
                        res = idx, new_res
        return res

    def get_points_col(self, col: int) -> Iterator[Tuple[int, int]]:
        """Yields all points of the gridded permutation in the column col."""
        return ((i, val) for i, (val, (x, _)) in enumerate(self) if x == col)

    def get_points_row(self, row: int) -> Iterator[Tuple[int, int]]:
        """Yields all points of the gridded permutation in the row."""
        return ((i, val) for i, (val, (_, y)) in enumerate(self) if y == row)

    def get_points_below_row(self, row: int) -> Iterator[Tuple[int, int]]:
        """Yields all points of the gridded permutation below the row."""
        return ((i, val) for i, (val, (_, y)) in enumerate(self) if y < row)

    def get_points_above_row(self, row: int) -> Iterator[Tuple[int, int]]:
        """Yields all points of the gridded permutation above the row."""
        return ((i, val) for i, (val, (_, y)) in enumerate(self) if y > row)

    def get_points_left_col(self, col) -> Iterator[Tuple[int, int]]:
        """Yields all points of the gridded permutation left of column col."""
        return ((i, val) for i, (val, (x, _)) in enumerate(self) if x < col)

    def get_subperm_left_col(self, col: int) -> "GriddedPerm":
        """Returns the gridded subpermutation of points left of column col."""
        gen1, gen2 = tee((i for i, _ in self.get_points_left_col(col)), 2)
        return type(self)(
            Perm.to_standard(self.patt[i] for i in gen1), (self.pos[i] for i in gen2)
        )

    def get_points_right_col(self, col: int) -> Iterator[Tuple[int, int]]:
        """Yields all points of the gridded permutation right of column col."""
        return ((i, val) for i, (val, (x, _)) in enumerate(self) if x > col)

    def get_gridded_perm_at_indices(self, indices: Iterable[int]) -> "GriddedPerm":
        """
        Returns the subgridded perm that contains only the point at the given
        indices. Indices must be sorted.
        """
        gen1, gen2 = tee(indices, 2)
        return type(self)(
            Perm.to_standard(self.patt[i] for i in gen1),
            (self.pos[i] for i in gen2),
        )

    def get_gridded_perm_in_cells(self, cells: Iterable[Cell]) -> "GriddedPerm":
        """Returns the subgridded permutation of points in cells."""
        cells = set(cells)
        return type(self)(
            Perm.to_standard(val for val, pos in self if pos in cells),
            (pos for pos in self.pos if pos in cells),
        )

    def get_bounding_box(self, cell: Cell) -> Tuple[int, int, int, int]:
        """Determines the indices and values of the gridded permutation in
        which a point that is being inserted into the cell given."""
        row = list(self.get_points_row(cell[1]))
        col = list(self.get_points_col(cell[0]))
        if not row:
            above = list(self.get_points_above_row(cell[1]))
            if not above:
                maxval = len(self)
            else:
                maxval = min(p[1] for p in above)
            below = list(self.get_points_below_row(cell[1]))
            if not below:
                minval = 0
            else:
                minval = max(p[1] for p in below) + 1
        else:
            maxval = max(p[1] for p in row) + 1
            minval = min(p[1] for p in row)
        if not col:
            right = list(self.get_points_right_col(cell[0]))
            if not right:
                maxdex = len(self)
            else:
                maxdex = min(p[0] for p in right)
            left = list(self.get_points_left_col(cell[0]))
            if not left:
                mindex = 0
            else:
                mindex = max(p[0] for p in left) + 1
        else:
            maxdex = max(p[0] for p in col) + 1
            mindex = min(p[0] for p in col)
        return (mindex, maxdex, minval, maxval)

    def insert_point(self, cell: Cell) -> Iterator["GriddedPerm"]:
        """Insert a new point into cell of the gridded perm, such that the
        point is added to the underlying pattern with the position at the cell.
        Yields all gridded perms where the point has been mixed into the points
        in the cell."""
        mindex, maxdex, minval, maxval = self.get_bounding_box(cell)
        for idx, val in product(range(mindex, maxdex + 1), range(minval, maxval + 1)):
            yield self.insert_specific_point(cell, idx, val)

    def insert_specific_point(self, cell: Cell, idx: int, val: int) -> "GriddedPerm":
        """Insert a point in the given cell with the given idx and val."""
        return type(self)(
            self._patt.insert(idx, val), self._pos[:idx] + (cell,) + self._pos[idx:]
        )

    def remove_point(self, index: int) -> "GriddedPerm":
        """Remove the point at index from the gridded permutation."""
        patt = Perm.to_standard(self.patt[:index] + self.patt[index + 1 :])
        pos = self.pos[:index] + self.pos[index + 1 :]
        return type(self)(patt, pos)

    def all_subperms(self, proper: bool = True) -> Iterator["GriddedPerm"]:
        """Yields all gridded subpermutations."""
        for r in range(len(self) if proper else len(self) + 1):
            for subidx in combinations(range(len(self)), r):
                yield type(self)(
                    Perm.to_standard(self._patt[i] for i in subidx),
                    (self._pos[i] for i in subidx),
                )

    def extend(self, c: int, r: int) -> Iterator["GriddedPerm"]:
        """Add n+1 to all possible positions in perm and all allowed positions given
        that placement."""
        n = len(self)
        if n == 0:
            yield from (
                GriddedPerm((0,), ((x, y),)) for x in range(c) for y in range(r)
            )
        else:
            min_y = max(y for _, (_, y) in self)
            yield from (
                self.insert_specific_point((x, y), index, n)
                for index in range(n + 1)
                for x in range(
                    self.pos[index - 1][0] if index > 0 else 0,
                    self.pos[index][0] + 1 if index < n else c,
                )
                for y in range(min_y, r)
            )

    def apply_map(self, cell_mapping: Callable[[Cell], Cell]) -> "GriddedPerm":
        """Map the coordinates to a new list of coordinates according to the
        cell_mapping given."""
        return type(self)(self._patt, [cell_mapping(cell) for cell in self._pos])

    def is_point_perm(self) -> bool:
        """Checks if the gridded permutation is of length 1."""
        return len(self) == 1

    def is_localized(self) -> bool:
        """Check if the gridded permutation occupies only a single cell."""
        return self.is_single_cell()

    def is_single_cell(self) -> bool:
        """Check if the gridded permutation occupies only a single cell."""
        return len(self._cells) == 1

    def is_single_row(self) -> bool:
        """Check if the gridded permutation occupies only a single row."""
        return len(set(y for (_, y) in self._cells)) == 1

    def is_empty(self) -> bool:
        """Check if the gridded permutation is the gridded permutation."""
        return not bool(self._patt)

    def is_interleaving(self) -> bool:
        """Check if the gridded permutation occupies two cells that are in the
        same row or column."""
        xs: Dict[int, int] = {}
        ys: Dict[int, int] = {}
        for x, y in self._cells:
            if xs.get(x, y) != y or ys.get(y, x) != x:
                return True
            xs[x] = y
            ys[y] = x
        return False

    def factors(self) -> List["GriddedPerm"]:
        """Return a list containing the factors of a gridded permutation.
        A factor is a sub gridded permutation that is isolated on its own rows
        and columns."""
        uf = UnionFind(len(self.pos))
        for j, (_, (x_r, y_r)) in enumerate(self):
            for i, (_, (x_l, y_l)) in enumerate(islice(self, j)):
                if x_l == x_r or y_l == y_r:
                    uf.unite(i, j)
        # Collect the connected factors of the cells
        all_factors: Dict[int, List[Cell]] = {}
        for i, cell in enumerate(self.pos):
            x = uf.find(i)
            if x in all_factors:
                all_factors[x].append(cell)
            else:
                all_factors[x] = [cell]
        factor_cells = list(set(cells) for cells in all_factors.values())
        return [self.get_gridded_perm_in_cells(comp) for comp in factor_cells]

    def compress(self) -> List[int]:
        """Compresses the gridded permutation into a list of integers.
        It starts with a list of the values in the permutation. The rest is
        the list of positions flattened."""
        return list(chain(self._patt, chain.from_iterable(self._pos)))

    @classmethod
    def decompress(cls, arr: array) -> "GriddedPerm":
        """Decompresses a list of integers in the form outputted by the
        compress method and constructs an Obstruction."""
        n, it = len(arr) // 3, iter(arr)
        return cls(
            Perm(next(it) for _ in range(n)), ((next(it), next(it)) for _ in range(n))
        )

    # Symmetries
    def reverse(self, transf: Callable[[Cell], Cell]) -> "GriddedPerm":
        """
        Reverses the tiling within its boundary. Every cell and obstruction
        gets flipped over the vertical middle axis."""
        return type(self)(self._patt.reverse(), reversed(list(map(transf, self._pos))))

    def complement(self, transf: Callable[[Cell], Cell]) -> "GriddedPerm":
        """Flip over the horizontal axis."""
        return type(self)(self._patt.complement(), map(transf, self._pos))

    def inverse(self, transf: Callable[[Cell], Cell]) -> "GriddedPerm":
        """Flip over the diagonal."""
        flipped = self._patt.inverse()
        pos = self._patt.inverse().apply(self._pos)
        return type(self)(flipped, map(transf, pos))

    def antidiagonal(self, transf: Callable[[Cell], Cell]) -> "GriddedPerm":
        """Flip over the anti-diagonal"""
        flipped = self._patt.flip_antidiagonal()
        pos = self._patt.rotate(-1).apply(self._pos)
        return type(self)(flipped, map(transf, pos))

    def rotate270(self, transf: Callable[[Cell], Cell]) -> "GriddedPerm":
        """Rotate 270 degrees"""
        rotated = self._patt.rotate(-1)
        pos = rotated.apply(self._pos)
        return type(self)(rotated, map(transf, pos))

    def rotate180(self, transf: Callable[[Cell], Cell]) -> "GriddedPerm":
        """Rotate 180 degrees"""
        return type(self)(self._patt.rotate(2), reversed(list(map(transf, self._pos))))

    def rotate90(self, transf: Callable[[Cell], Cell]) -> "GriddedPerm":
        """Rotate 90 degrees"""
        return type(self)(
            self._patt.rotate(), map(transf, self._patt.inverse().apply(self._pos))
        )

    def column_reverse(self, column: int) -> "GriddedPerm":
        """Reverse the part of the gridded perm that belongs to a given column."""
        parts: Tuple[List[Tuple[int, Tuple[int, int]]], ...] = ([], [], [])
        for v, (x, y) in self:
            parts[(x >= column) + (x > column)].append((v, (x, y)))
        return type(self)(*zip(*chain(parts[0], reversed(parts[1]), parts[2])))

    def row_complement(self, row: int) -> "GriddedPerm":
        """Replace the part of the gridded perm that belongs to a row with its
        complement."""
        indices, vals = set(), []
        for i, (val, (_, y)) in enumerate(self):
            if y == row:
                indices.add(i)
                vals.append(val)
        st = Perm.to_standard(vals)
        unstandardized = dict(zip(st, vals))
        in_row = (unstandardized[val] for val in st.complement())
        return type(self)(
            Perm(
                next(in_row) if i in indices else val for i, val in enumerate(self.patt)
            ),
            self.pos,
        )

    def permute_columns(self, perm: Iterable[int]) -> "GriddedPerm":
        """Given an initial state of columns 12...n, permute them using the provided
        permutation.
        """
        if not isinstance(perm, Perm):
            perm = Perm(perm)
        assert len(perm) > max(x for x, y in self.pos)
        cols: List[List[Tuple[int, int]]] = [[] for _ in range(len(perm))]
        for v, (x, y) in self:
            cols[x].append((v, y))
        patt, positions = [], []
        for x, col in enumerate(perm.apply(cols)):
            for v, y in col:
                patt.append(v)
                positions.append((x, y))
        return type(self)(patt, positions)

    def permute_rows(self, perm: Iterable[int]) -> "GriddedPerm":
        """Given an initial state of rows 12...n, permute them using the provided
        permutation.
        """
        if not isinstance(perm, Perm):
            perm = Perm(perm)
        assert len(perm) > max(x for x, y in self.pos)
        back_map = dict(zip(perm, range(len(perm))))
        positions = [(x, back_map[y]) for x, y in self.pos]
        occ: List[List[int]] = [[] for _ in range(len(perm))]
        for val, (_, y) in zip(self.patt, positions):
            occ[y].append(val)
        offset, val_map = 0, {}
        for lis in occ:
            for a, b in zip(lis, Perm.to_standard(lis)):
                val_map[a] = b + offset
            offset += len(lis)
        return type(self)((val_map[val] for val in self.patt), positions)

    def apply_perm_map_to_cell(
        self, perm_mapping: Callable[[Perm], Perm], cell: Cell
    ) -> "GriddedPerm":
        """Apply a permutation map to the subperm within a cell."""
        subperm = [val for val, pos in self if pos == cell]
        st = Perm.to_standard(subperm)
        back_map = dict(zip(st, subperm))
        new_subperm = (back_map[val] for val in perm_mapping(st))
        return type(self)(
            (next(new_subperm) if pos == cell else val for val, pos in self), self.pos
        )

    def to_jsonable(self) -> dict:
        """Returns a dictionary object which is JSON serializable representing
        a GriddedPerm."""
        return {"patt": self._patt, "pos": self._pos}

    @classmethod
    def from_json(cls, jsonstr: str) -> "GriddedPerm":
        """Returns a GriddedPerm object from JSON string."""
        jsondict = json.loads(jsonstr)
        return cls.from_dict(jsondict)

    @classmethod
    def from_dict(cls, jsondict: dict) -> "GriddedPerm":
        """Returns a GriddedPerm object from a dictionary loaded from a JSON
        serialized GriddedPerm object."""
        return cls(jsondict["patt"], map(tuple, jsondict["pos"]))  # type: ignore

    @property
    def patt(self) -> Perm:
        return self._patt

    @property
    def pos(self) -> Tuple[Cell, ...]:
        return self._pos

    def ascii_plot(self) -> str:
        max_x = max(cell[0] for cell in self.pos)
        max_y = max(cell[1] for cell in self.pos)
        res = ""

        def points_in_col(i):
            return sum(1 for cell in self.pos if cell[0] == i)

        def points_in_row(j):
            return sum(1 for cell in self.pos if cell[1] == j)

        row_boundary = (
            "+" + "+".join("-" * points_in_col(i) for i in range(max_x + 1)) + "+"
        )
        col_boundary = (
            "|" + "|".join(" " * points_in_col(i) for i in range(max_x + 1)) + "|"
        )

        for j in range(max_y, -1, -1):
            res += (
                "\n".join(
                    [row_boundary] + [col_boundary for i in range(points_in_row(j))]
                )
                + "\n"
            )
        res += row_boundary

        for (idx, val) in enumerate(self.patt):
            x, y = self.pos[idx]
            # insert into this spot:
            # (idx + x + 1) is the horizontal index. idx is points to left, and
            #               x + 1 counts number of col boundaries to left
            # (len(self) + max_y) - (val + y) is vertical index.
            #               val is points below, and y + 1 counts number of - below
            insert = (idx + x + 1) + ((len(self) + max_y) - (val + y)) * (
                len(col_boundary) + 1
            )
            res = res[:insert] + "●" + res[insert + 1 :]
        return res

    def to_tex(self, one_line_notation: bool = True, one_based: bool = True) -> str:
        """Return the GriddedPerm in LaTeX math mode."""
        if one_line_notation:
            return "".join(f"{val + one_based}^{{({x},{y})}}" for val, (x, y) in self)
        perm = (
            "".join(
                (f"({val + one_based})" for val in self.patt)
                if len(self) > 9
                else (f"{val + one_based}" for val in self.patt)
            )
            if one_based
            else f"{self.patt}"
        )
        return f"({perm}, {self.pos})"

    def _get_plot_pos(self) -> Tuple[Dict[int, Tuple[float, float]], Tuple[int, int]]:
        """Helper that positions visualization points of GriddedPerm."""
        dim = (1 + max(x for _, (x, _) in self), 1 + max(y for _, (_, y) in self))
        col_to_val: List[List[Tuple[int, int]]] = [[] for i in range(dim[0])]
        row_to_val: List[List[Tuple[int, int]]] = [[] for i in range(dim[1])]
        val_to_pos = {}
        for val, (x, y) in self:
            col_to_val[x].append((val, len(col_to_val[x])))
            row_to_val[y].append((val, len(row_to_val[y])))
            val_to_pos[val] = (x, y)
        return {
            val: (
                next(
                    (
                        x + (k + 1) * (1 / (len(col_to_val[x]) + 1))
                        for j, k in sorted(col_to_val[x])
                        if j == val
                    )
                ),
                next(
                    (
                        y + (i + 1) * (1 / (len(row_to_val[y]) + 1))
                        for i, (j, _) in enumerate(sorted(row_to_val[y]))
                        if j == val
                    )
                ),
            )
            for val, (x, y) in val_to_pos.items()
        }, dim

    def to_tikz(self, tab: str = "  ") -> str:
        """Return the tikz code to plot the GriddedPerm."""
        val_to_pos, (max_x, max_y) = self._get_plot_pos()
        return "".join(
            (
                "\\begin{tikzpicture}\n",
                f"{tab}\\def\\xs{{1.0}}\n",
                f"{tab}\\def\\ys{{1.0}}\n",
                f"{tab}\\def\\ps{{1.0}}\n",
                (
                    f"{tab}\\draw (0,0) grid[xscale=\\xs,yscale=\\ys] "
                    f"({max_x}, {max_y});\n"
                ),
                "\n".join(
                    (
                        f"{tab}\\coordinate (p{i}) at "
                        f"({val_to_pos[v][0]}*\\xs,{val_to_pos[v][1]}*\\ys);"
                    )
                    for i, v in enumerate(self.patt)
                ),
                f'\n{tab}\\draw {"--".join(f"(p{i})" for i in range(len(self)))};\n',
                "\n".join(
                    f"{tab}\\fill (p{i}) circle (0.1*\\ps);" for i in range(len(self))
                ),
                "\n\\end{tikzpicture}",
            )
        )

    def to_svg(self, image_scale: float = 10.0) -> str:
        """Return the svg code to plot the GriddedPerm."""
        i_scale = int(image_scale * 10)
        val_to_pos, (m_x, m_y) = self._get_plot_pos()
        points = " ".join(
            f"{x * 10},{(m_y - y) * 10}"
            for x, y in (val_to_pos[val] for val in self.patt)
        )
        return "".join(
            [
                (
                    '<svg xmlns="http://www.w3.org/2000/svg" version="1.1" width='
                    f'"{m_x * i_scale}" height="{m_y * i_scale}" '
                    f'viewBox="0 0 {m_x*10} {m_y*10}">\n'
                ),
                "\n".join(
                    (
                        f'<line x1="{i*10}" y1="0" x2="{i*10}" y2="{m_y*10}" '
                        'style="stroke:rgb(0,0,0);stroke-width:0.5" />'
                    )
                    for i in range(m_x + 1)
                ),
                "\n",
                "\n".join(
                    (
                        f'<line x1="0" y1="{i*10}" x2="{m_x*10}" y2="{i*10}" '
                        'style="stroke:rgb(0,0,0);stroke-width:0.5" />'
                    )
                    for i in range(m_y + 1)
                ),
                "\n",
                (
                    f'<polyline points="{points}" fill="none" '
                    'stroke="black" stroke-width="0.65" />\n'
                ),
                "\n".join(
                    (
                        f'<circle cx="{(10 * x):.4f}" cy="{(10 * (m_y - y)):.4f}" '
                        'r="0.5" stroke="black" fill="rgb(0,0,0)" />'
                    )
                    for x, y in (val_to_pos[val] for val in self.patt)
                ),
                "\n</svg>",
            ]
        )

    def show(self, scale: float = 10.0) -> None:
        """Open a browser tab and display GriddedPerm graphically. Image can be
        enlarged with scale parameter
        """
        HTMLViewer.open_svg(self.to_svg(image_scale=scale))

    def __len__(self) -> int:
        return len(self._patt)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({tuple(self._patt)!r}, {self.pos})"

    def __str__(self) -> str:
        return f"{self._patt}: {', '.join(str(c) for c in self.pos)}"

    def __hash__(self) -> int:
        return hash(self._patt) ^ hash(self._pos)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return False
        return self._patt == other.patt and self._pos == other.pos

    def __lt__(self, other: "GriddedPerm") -> bool:
        return (self._patt, self._pos) < (other.patt, other.pos)

    def __contains__(self, other: "GriddedPerm") -> bool:
        return next((True for _ in other.occurrences_in(self)), False)

    def __iter__(self) -> Iterator[Tuple[int, Cell]]:
        return zip(self.patt, self.pos)



import pytest

from permuta import Perm
from permuta.misc import DIR_EAST, DIR_NORTH, DIR_SOUTH, DIR_WEST
from tilings import GriddedPerm


@pytest.fixture
def simpleob():
    return GriddedPerm((1, 0, 3, 2), ((0, 0), (0, 0), (2, 2), (2, 1)))


@pytest.fixture
def singlecellob():
    return GriddedPerm.single_cell((1, 0, 3, 2), (2, 2))


@pytest.fixture
def everycellob():
    return GriddedPerm(
        Perm((0, 3, 6, 1, 4, 7, 2, 5, 8)),
        ((0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2), (2, 0), (2, 1), (2, 2)),
    )


@pytest.fixture
def typicalob():
    return GriddedPerm((1, 0, 2, 4, 3), ((0, 0), (0, 0), (1, 0), (1, 1), (1, 1)))


@pytest.fixture
def isolatedob():
    return GriddedPerm((0, 1, 2), ((0, 0), (1, 1), (2, 2)))


def test_from_iterable():
    pos = ((0, 0), (1, 1), (2, 1))
    gp = GriddedPerm(Perm((0, 1, 2)), pos)
    assert GriddedPerm((0, 1, 2), pos) == gp
    assert GriddedPerm([0, 1, 2], pos) == gp
    assert GriddedPerm(range(3), pos) == gp
    assert GriddedPerm((i for i in range(3)), pos) == gp
    assert GriddedPerm.single_cell((0, 1), (0, 0)) == GriddedPerm(
        Perm((0, 1)), ((0, 0), (0, 0))
    )
    assert GriddedPerm(
        (0, 2, 1, 3, 4), ((0, 0), (0, 1), (1, 0), (1, 1), (1, 1))
    )._cells == {
        (0, 0),
        (0, 1),
        (1, 0),
        (1, 1),
    }


def test_single_cell():
    gp = GriddedPerm.single_cell((0, 1, 2), (2, 3))
    assert gp.patt == Perm((0, 1, 2))
    assert gp.pos == ((2, 3), (2, 3), (2, 3))

    gp = GriddedPerm.single_cell((0,), (45, 64))
    assert gp.patt == Perm((0,))
    assert gp.pos == ((45, 64),)


def test_empty_perm():
    gp = GriddedPerm.empty_perm()
    assert gp.patt == Perm(())
    assert gp.pos == ()
    assert len(gp._cells) == 0
    assert GriddedPerm() == gp == GriddedPerm((), ())


def test_contradictory(simpleob, singlecellob, everycellob):
    assert not simpleob.contradictory()
    assert not singlecellob.contradictory()
    assert not everycellob.contradictory()
    gp = GriddedPerm((0, 1), [(0, 1), (1, 0)])
    assert gp.contradictory()
    gp = GriddedPerm((0, 1), [(1, 0), (0, 0)])
    assert gp.contradictory()
    gp = GriddedPerm((1, 0), [(0, 0), (1, 1)])
    assert gp.contradictory()


def test_occupies(simpleob):
    assert simpleob.occupies((0, 0))
    assert simpleob.occupies((2, 1))
    assert simpleob.occupies((2, 2))

    assert not simpleob.occupies((0, 1))
    assert not simpleob.occupies((3, 1))
    assert not simpleob.occupies((2, 0))


def test_occurrences_in(simpleob):
    ob = GriddedPerm((0, 2, 1), ((0, 0), (2, 2), (2, 1)))
    assert list(ob.occurrences_in(simpleob)) == [(0, 2, 3), (1, 2, 3)]
    assert ob.occurs_in(simpleob)

    ob = GriddedPerm((1, 0, 2), ((0, 0), (0, 0), (2, 1)))
    assert list(ob.occurrences_in(simpleob)) == [(0, 1, 3)]
    assert ob.occurs_in(simpleob)

    ob = GriddedPerm((1, 0, 2), ((0, 0), (0, 0), (2, 2)))
    assert list(ob.occurrences_in(simpleob)) == [(0, 1, 2)]
    assert ob.occurs_in(simpleob)

    ob = GriddedPerm((0, 1, 2), ((0, 0), (2, 2), (2, 1)))
    assert not any(ob.occurrences_in(simpleob))
    assert not ob.occurs_in(simpleob)

    ob = GriddedPerm((1, 0, 2), ((0, 0), (2, 2), (2, 2)))
    assert not any(ob.occurrences_in(simpleob))
    assert not ob.occurs_in(simpleob)

    ob = GriddedPerm(Perm(tuple()), tuple())
    assert ob in ob


def test_avoids():
    gp = GriddedPerm(
        Perm((0, 5, 1, 4, 2, 3, 6)),
        ((0, 0), (0, 1), (0, 0), (1, 1), (1, 1), (1, 1), (2, 2)),
    )
    assert gp.avoids(GriddedPerm((0, 2, 1), ((0, 0),) * 3))
    assert gp.avoids(
        GriddedPerm((0, 2, 1), ((0, 0),) * 3),
        GriddedPerm((0, 2, 1), ((1, 1), (1, 1), (2, 2))),
    )


def test_contains():
    gp = GriddedPerm(
        Perm((0, 5, 1, 4, 2, 3, 6)),
        ((0, 0), (0, 1), (0, 0), (1, 1), (1, 1), (1, 1), (2, 2)),
    )
    assert gp.contains(GriddedPerm())
    assert all(gp.contains(GriddedPerm((0,), (pos,))) for pos in gp._cells)
    assert gp.contains(GriddedPerm((0, 1, 2), ((0, 0), (0, 0), (2, 2))))
    assert gp.contains(GriddedPerm((2, 0, 1), ((1, 1),) * 3))
    assert gp.contains(
        GriddedPerm((0, 2, 1), ((0, 0),) * 3),
        GriddedPerm((0, 2, 1), ((1, 1), (1, 1), (2, 2))),
        GriddedPerm((1, 0, 2), ((1, 1), (1, 1), (2, 2))),
    )
    assert GriddedPerm().contains(GriddedPerm())


def test_remove_cells(simpleob, everycellob):
    assert simpleob.remove_cells([(0, 0)]) == GriddedPerm((1, 0), ((2, 2), (2, 1)))
    assert simpleob.remove_cells([(0, 0), (2, 2)]) == GriddedPerm((0,), ((2, 1),))
    assert simpleob.remove_cells([(0, 1), (1, 2)]) == simpleob
    gen = (cell for cell in ((0, 1), (1, 0)))
    assert everycellob.remove_cells(gen) == GriddedPerm(
        Perm((0, 4, 2, 5, 1, 3, 6)),
        ((0, 0), (0, 2), (1, 1), (1, 2), (2, 0), (2, 1), (2, 2)),
    )


def test_points_in_cell(simpleob):
    assert list(simpleob.points_in_cell((0, 0))) == [0, 1]
    assert list(simpleob.points_in_cell((0, 1))) == []


def test_isolated_cells(simpleob, isolatedob):
    assert list(simpleob.isolated_cells()) == []
    assert list(isolatedob.isolated_cells()) == [(0, 0), (1, 1), (2, 2)]


def test_is_isolated(simpleob, isolatedob):
    assert simpleob.is_isolated([2, 3])
    assert simpleob.is_isolated([0, 1])
    assert not simpleob.is_isolated([0])
    assert not simpleob.is_isolated([2])
    assert not simpleob.is_isolated([1, 3])
    assert isolatedob.is_isolated([0, 1, 2])
    assert isolatedob.is_isolated([0, 1])
    assert isolatedob.is_isolated([0, 2])
    assert isolatedob.is_isolated([1, 2])
    assert isolatedob.is_isolated([0])
    assert isolatedob.is_isolated([1])
    assert isolatedob.is_isolated([2])


def test_forced_point_index(singlecellob):
    with pytest.raises(ValueError):
        singlecellob.forced_point_index((1, 1), DIR_SOUTH)
    assert singlecellob.forced_point_index((2, 2), DIR_WEST) == 0
    assert singlecellob.forced_point_index((2, 2), DIR_SOUTH) == 1
    assert singlecellob.forced_point_index((2, 2), DIR_NORTH) == 2
    assert singlecellob.forced_point_index((2, 2), DIR_EAST) == 3


def test_get_rowcol(everycellob, simpleob, typicalob):
    assert list(simpleob.get_points_col(0)) == [(0, 1), (1, 0)]
    assert list(simpleob.get_points_col(1)) == []
    assert list(simpleob.get_points_col(2)) == [(2, 3), (3, 2)]

    assert list(typicalob.get_points_col(0)) == [(0, 1), (1, 0)]
    assert list(typicalob.get_points_col(1)) == [(2, 2), (3, 4), (4, 3)]
    assert list(typicalob.get_points_col(2)) == []

    assert list(simpleob.get_points_row(0)) == [(0, 1), (1, 0)]
    assert list(simpleob.get_points_row(1)) == [(3, 2)]
    assert list(simpleob.get_points_row(2)) == [(2, 3)]
    assert list(simpleob.get_points_row(3)) == []

    assert list(typicalob.get_points_row(0)) == [(0, 1), (1, 0), (2, 2)]
    assert list(typicalob.get_points_row(1)) == [(3, 4), (4, 3)]
    assert list(typicalob.get_points_row(2)) == []

    assert list(everycellob.get_points_row(1)) == [(1, 3), (4, 4), (7, 5)]
    assert list(everycellob.get_points_col(1)) == [(3, 1), (4, 4), (5, 7)]
    assert list(everycellob.get_points_above_row(2)) == []
    assert list(everycellob.get_points_below_row(2)) == [
        (0, 0),
        (1, 3),
        (3, 1),
        (4, 4),
        (6, 2),
        (7, 5),
    ]
    assert list(everycellob.get_points_right_col(1)) == [(6, 2), (7, 5), (8, 8)]
    assert list(everycellob.get_points_left_col(1)) == [(0, 0), (1, 3), (2, 6)]


def test_get_subperm_left_col(everycellob, simpleob, typicalob):
    assert everycellob.get_subperm_left_col(0) == GriddedPerm((), ())
    assert everycellob.get_subperm_left_col(1) == GriddedPerm(
        (0, 1, 2), ((0, 0), (0, 1), (0, 2))
    )
    assert everycellob.get_subperm_left_col(2) == GriddedPerm(
        (0, 2, 4, 1, 3, 5), ((0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2))
    )
    assert everycellob.get_subperm_left_col(3) == everycellob

    assert simpleob.get_subperm_left_col(1) == GriddedPerm((1, 0), ((0, 0), (0, 0)))
    assert simpleob.get_subperm_left_col(10) == simpleob

    assert typicalob.get_subperm_left_col(1) == GriddedPerm((1, 0), ((0, 0), (0, 0)))
    assert typicalob.get_subperm_left_col(100) == typicalob


def test_get_bounding_box(typicalob):
    assert typicalob.get_bounding_box((1, 0)) == (2, 5, 0, 3)
    assert typicalob.get_bounding_box((2, 2)) == (5, 5, 5, 5)


def test_is_point_perm(typicalob, singlecellob):
    assert not typicalob.is_point_perm()
    assert not singlecellob.is_point_perm()
    assert GriddedPerm((0,), ((0, 0),)).is_point_perm()
    assert GriddedPerm((0,), ((3, 2),)).is_point_perm()
    assert GriddedPerm((0,), ((100, 10),)).is_point_perm()


def test_is_single_cell(typicalob, simpleob, singlecellob):
    assert not typicalob.is_single_cell()
    assert not simpleob.is_single_cell()
    assert singlecellob.is_single_cell()


def test_insert_point():
    ob = GriddedPerm(
        (0, 4, 1, 2, 3, 5), [(0, 0), (0, 1), (0, 0), (1, 1), (1, 1), (2, 2)]
    )
    assert list(ob.insert_point((1, 1))) == [
        GriddedPerm(
            Perm((0, 5, 1, 2, 3, 4, 6)),
            ((0, 0), (0, 1), (0, 0), (1, 1), (1, 1), (1, 1), (2, 2)),
        ),
        GriddedPerm(
            Perm((0, 5, 1, 3, 2, 4, 6)),
            ((0, 0), (0, 1), (0, 0), (1, 1), (1, 1), (1, 1), (2, 2)),
        ),
        GriddedPerm(
            Perm((0, 5, 1, 4, 2, 3, 6)),
            ((0, 0), (0, 1), (0, 0), (1, 1), (1, 1), (1, 1), (2, 2)),
        ),
        GriddedPerm(
            Perm((0, 4, 1, 5, 2, 3, 6)),
            ((0, 0), (0, 1), (0, 0), (1, 1), (1, 1), (1, 1), (2, 2)),
        ),
        GriddedPerm(
            Perm((0, 5, 1, 3, 2, 4, 6)),
            ((0, 0), (0, 1), (0, 0), (1, 1), (1, 1), (1, 1), (2, 2)),
        ),
        GriddedPerm(
            Perm((0, 5, 1, 2, 3, 4, 6)),
            ((0, 0), (0, 1), (0, 0), (1, 1), (1, 1), (1, 1), (2, 2)),
        ),
        GriddedPerm(
            Perm((0, 5, 1, 2, 4, 3, 6)),
            ((0, 0), (0, 1), (0, 0), (1, 1), (1, 1), (1, 1), (2, 2)),
        ),
        GriddedPerm(
            Perm((0, 4, 1, 2, 5, 3, 6)),
            ((0, 0), (0, 1), (0, 0), (1, 1), (1, 1), (1, 1), (2, 2)),
        ),
        GriddedPerm(
            Perm((0, 5, 1, 3, 4, 2, 6)),
            ((0, 0), (0, 1), (0, 0), (1, 1), (1, 1), (1, 1), (2, 2)),
        ),
        GriddedPerm(
            Perm((0, 5, 1, 2, 4, 3, 6)),
            ((0, 0), (0, 1), (0, 0), (1, 1), (1, 1), (1, 1), (2, 2)),
        ),
        GriddedPerm(
            Perm((0, 5, 1, 2, 3, 4, 6)),
            ((0, 0), (0, 1), (0, 0), (1, 1), (1, 1), (1, 1), (2, 2)),
        ),
        GriddedPerm(
            Perm((0, 4, 1, 2, 3, 5, 6)),
            ((0, 0), (0, 1), (0, 0), (1, 1), (1, 1), (1, 1), (2, 2)),
        ),
    ]


def test_all_subperms(simpleob):
    assert list(sorted(simpleob.all_subperms())) == [
        GriddedPerm((), ()),
        GriddedPerm((0,), ((0, 0),)),
        GriddedPerm((0,), ((0, 0),)),
        GriddedPerm((0,), ((2, 1),)),
        GriddedPerm((0,), ((2, 2),)),
        GriddedPerm((0, 1), ((0, 0), (2, 1))),
        GriddedPerm((0, 1), ((0, 0), (2, 1))),
        GriddedPerm((0, 1), ((0, 0), (2, 2))),
        GriddedPerm((0, 1), ((0, 0), (2, 2))),
        GriddedPerm((1, 0), ((0, 0), (0, 0))),
        GriddedPerm((1, 0), ((2, 2), (2, 1))),
        GriddedPerm((0, 2, 1), ((0, 0), (2, 2), (2, 1))),
        GriddedPerm((0, 2, 1), ((0, 0), (2, 2), (2, 1))),
        GriddedPerm((1, 0, 2), ((0, 0), (0, 0), (2, 1))),
        GriddedPerm((1, 0, 2), ((0, 0), (0, 0), (2, 2))),
    ]
    assert list(sorted(simpleob.all_subperms(proper=False))) == [
        GriddedPerm((), ()),
        GriddedPerm((0,), ((0, 0),)),
        GriddedPerm((0,), ((0, 0),)),
        GriddedPerm((0,), ((2, 1),)),
        GriddedPerm((0,), ((2, 2),)),
        GriddedPerm((0, 1), ((0, 0), (2, 1))),
        GriddedPerm((0, 1), ((0, 0), (2, 1))),
        GriddedPerm((0, 1), ((0, 0), (2, 2))),
        GriddedPerm((0, 1), ((0, 0), (2, 2))),
        GriddedPerm((1, 0), ((0, 0), (0, 0))),
        GriddedPerm((1, 0), ((2, 2), (2, 1))),
        GriddedPerm((0, 2, 1), ((0, 0), (2, 2), (2, 1))),
        GriddedPerm((0, 2, 1), ((0, 0), (2, 2), (2, 1))),
        GriddedPerm((1, 0, 2), ((0, 0), (0, 0), (2, 1))),
        GriddedPerm((1, 0, 2), ((0, 0), (0, 0), (2, 2))),
        GriddedPerm((1, 0, 3, 2), ((0, 0), (0, 0), (2, 2), (2, 1))),
    ]


def test_remove_point(typicalob, simpleob, singlecellob):
    assert simpleob.remove_point(0) == GriddedPerm((0, 2, 1), [(0, 0), (2, 2), (2, 1)])
    assert simpleob.remove_point(1) == GriddedPerm((0, 2, 1), [(0, 0), (2, 2), (2, 1)])
    assert simpleob.remove_point(2) == GriddedPerm((1, 0, 2), [(0, 0), (0, 0), (2, 1)])
    assert simpleob.remove_point(3) == GriddedPerm((1, 0, 2), [(0, 0), (0, 0), (2, 2)])

    assert typicalob.remove_point(0) == GriddedPerm(
        (0, 1, 3, 2), [(0, 0), (1, 0), (1, 1), (1, 1)]
    )
    assert typicalob.remove_point(1) == GriddedPerm(
        (0, 1, 3, 2), [(0, 0), (1, 0), (1, 1), (1, 1)]
    )
    assert typicalob.remove_point(2) == GriddedPerm(
        (1, 0, 3, 2), [(0, 0), (0, 0), (1, 1), (1, 1)]
    )
    assert typicalob.remove_point(3) == GriddedPerm(
        (1, 0, 2, 3), [(0, 0), (0, 0), (1, 0), (1, 1)]
    )
    assert typicalob.remove_point(4) == GriddedPerm(
        (1, 0, 2, 3), [(0, 0), (0, 0), (1, 0), (1, 1)]
    )

    assert singlecellob.remove_point(0) == GriddedPerm(
        (0, 2, 1), [(2, 2), (2, 2), (2, 2)]
    )
    assert singlecellob.remove_point(1) == GriddedPerm(
        (0, 2, 1), [(2, 2), (2, 2), (2, 2)]
    )
    assert singlecellob.remove_point(2) == GriddedPerm(
        (1, 0, 2), [(2, 2), (2, 2), (2, 2)]
    )
    assert singlecellob.remove_point(3) == GriddedPerm(
        (1, 0, 2), [(2, 2), (2, 2), (2, 2)]
    )


def test_apply_map(typicalob, simpleob):
    def cell_mapping(x):
        return (x[0] + 1, x[1] + 2)

    assert typicalob.apply_map(cell_mapping) == GriddedPerm(
        (1, 0, 2, 4, 3), ((1, 2), (1, 2), (2, 2), (2, 3), (2, 3))
    )
    assert simpleob.apply_map(cell_mapping) == GriddedPerm(
        (1, 0, 3, 2), ((1, 2), (1, 2), (3, 4), (3, 3))
    )


def test_compression(simpleob, singlecellob, everycellob, typicalob, isolatedob):
    assert simpleob == GriddedPerm.decompress(simpleob.compress())
    assert singlecellob == GriddedPerm.decompress(singlecellob.compress())
    assert everycellob == GriddedPerm.decompress(everycellob.compress())
    assert typicalob == GriddedPerm.decompress(typicalob.compress())
    assert isolatedob == GriddedPerm.decompress(isolatedob.compress())


def test_plot_helper():
    gp = GriddedPerm(
        Perm((0, 3, 6, 1, 4, 7, 2, 5, 8)),
        ((0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2), (2, 0), (2, 1), (2, 2)),
    )
    pos, dim = gp._get_plot_pos()
    assert dim == (3, 3)
    assert pos[0] == (0.25, 0.25)
    assert pos[3] == (0.5, 1.25)
    assert pos[6] == (0.75, 2.25)
    assert pos[1] == (1.25, 0.5)
    assert pos[4] == (1.5, 1.5)
    assert pos[7] == (1.75, 2.5)
    assert pos[2] == (2.25, 0.75)
    assert pos[5] == (2.5, 1.75)
    assert pos[8] == (2.75, 2.75)


def test_permute_columns():
    gp1 = GriddedPerm(
        (5, 1, 4, 2, 7, 3, 6, 0),
        ((0, 1), (0, 0), (1, 1), (1, 0), (1, 2), (2, 1), (3, 2), (3, 0)),
    )
    perm1 = (3, 2, 0, 1)
    expected1 = GriddedPerm(
        (6, 0, 3, 5, 1, 4, 2, 7),
        ((0, 2), (0, 0), (1, 1), (2, 1), (2, 0), (3, 1), (3, 0), (3, 2)),
    )

    gp2 = GriddedPerm(
        (0,),
        ((0, 0),),
    )
    perm2 = (1, 2, 0)
    expected2 = GriddedPerm(
        (0,),
        ((2, 0),),
    )

    assert gp1.permute_columns(perm1) == expected1
    assert gp2.permute_columns(perm2) == expected2


def test_column_reverse():
    assert GriddedPerm(
        (4, 1, 3, 0, 5, 2), ((0, 1), (0, 0), (1, 1), (1, 0), (1, 2), (2, 1))
    ).column_reverse(1) == GriddedPerm(
        (4, 1, 5, 0, 3, 2), ((0, 1), (0, 0), (1, 2), (1, 0), (1, 1), (2, 1))
    )


def test_row_complement():
    assert GriddedPerm(
        (2, 0, 4, 5, 8, 1, 7, 6, 3),
        ((0, 1), (0, 0), (0, 1), (0, 1), (0, 2), (1, 0), (1, 2), (1, 1), (1, 1)),
    ).row_complement(1) == GriddedPerm(
        (6, 0, 4, 3, 8, 1, 7, 2, 5),
        ((0, 1), (0, 0), (0, 1), (0, 1), (0, 2), (1, 0), (1, 2), (1, 1), (1, 1)),
    )

    for gp in (
        GriddedPerm((1, 0), ((2, 2), (2, 1))),
        GriddedPerm(
            (5, 1, 4, 2, 7, 3, 6, 0),
            ((0, 1), (0, 0), (1, 1), (1, 0), (1, 2), (2, 1), (3, 2), (3, 0)),
        ),
    ):
        for i in range(3):
            assert gp.rotate90(lambda z: (z[1], 3 - z[0] - 1)).column_reverse(
                i
            ).rotate270(lambda z: (3 - z[1] - 1, z[0])) == gp.row_complement(i)


def test_permute_rows():
    assert GriddedPerm(
        Perm((0, 2, 1, 4, 3)), ((0, 0), (0, 1), (0, 0), (1, 2), (2, 1))
    ).permute_rows((2, 0, 1)) == GriddedPerm(
        Perm((1, 3, 2, 0, 4)), ((0, 1), (0, 2), (0, 1), (1, 0), (2, 2))
    )

    assert GriddedPerm((0,), ((0, 0),),).permute_rows((1, 2, 0)) == GriddedPerm(
        (0,),
        ((0, 2),),
    )


def test_to_jsonable():
    assert GriddedPerm(
        (5, 1, 4, 2, 7, 3, 6, 0),
        ((0, 1), (0, 0), (1, 1), (1, 0), (1, 2), (2, 1), (3, 2), (3, 0)),
    ).to_jsonable() == {
        "patt": Perm((5, 1, 4, 2, 7, 3, 6, 0)),
        "pos": ((0, 1), (0, 0), (1, 1), (1, 0), (1, 2), (2, 1), (3, 2), (3, 0)),
    }
    assert GriddedPerm((), ()).to_jsonable() == {"patt": Perm(()), "pos": ()}


def test_apply_perm_map_to_cell():
    def do_nothing_for_empty_perm(perm):
        assert perm == ()
        return perm

    gp = GriddedPerm((1, 0, 2, 4, 3), ((1, 2), (1, 2), (2, 2), (2, 3), (2, 3)))
    assert gp.apply_perm_map_to_cell(do_nothing_for_empty_perm, (0, 0)) == gp

    assert (
        GriddedPerm().apply_perm_map_to_cell(do_nothing_for_empty_perm, (0, 0))
        == GriddedPerm()
    )

    pos = [(0, 0), (0, 1), (0, 0), (0, 1), (0, 1), (0, 2), (1, 1)]
    gp = GriddedPerm((0, 4, 1, 2, 3, 6, 5), pos)

    assert gp.apply_perm_map_to_cell(lambda p: p.complement(), (0, 1)) == GriddedPerm(
        (0, 2, 1, 4, 3, 6, 5), pos
    )
    assert gp.apply_perm_map_to_cell(lambda p: p.inverse(), (0, 1)) == GriddedPerm(
        (0, 3, 1, 4, 2, 6, 5), pos
    )
    assert gp.apply_perm_map_to_cell(lambda p: p.reverse(), (0, 1)) == GriddedPerm(
        (0, 3, 1, 2, 4, 6, 5), pos
    )


def test_is_interleaving():
    assert GriddedPerm(
        (5, 1, 4, 2, 7, 3, 6, 0),
        ((0, 1), (0, 0), (1, 1), (1, 0), (1, 2), (2, 1), (3, 2), (3, 0)),
    ).is_interleaving()
    assert not GriddedPerm(
        Perm((0, 1, 2, 3, 4, 5)), ((i, i) for i in range(6))
    ).is_interleaving()
    assert not GriddedPerm().is_interleaving()
    assert not GriddedPerm((0,), ((0, 0),)).is_interleaving()
    assert GriddedPerm(
        Perm((0, 6, 1, 2, 3, 4, 5)),
        [(0, 0), (0, 20), (1, 1), (2, 2), (3, 3), (4, 4), (5, 5)],
    ).is_interleaving()
    assert GriddedPerm(
        Perm((0, 2, 3, 4, 5, 6, 1)),
        [(0, 0), (1, 1), (2, 2), (3, 3), (4, 4), (5, 5), (6, 0)],
    ).is_interleaving()


def test_extend():
    gp = GriddedPerm((1, 0), ((0, 1), (1, 0)))
    assert set(gp.extend(2, 2)) == {
        GriddedPerm((2, 1, 0), ((0, 1), (0, 1), (1, 0))),
        GriddedPerm((1, 2, 0), ((0, 1), (0, 1), (1, 0))),
        GriddedPerm((1, 2, 0), ((0, 1), (1, 1), (1, 0))),
        GriddedPerm((1, 0, 2), ((0, 1), (1, 0), (1, 1))),
    }



import gc

import pytest
import sympy

from comb_spec_searcher import CombinatorialSpecification
from comb_spec_searcher.rule_db import RuleDBForest
from comb_spec_searcher.strategies import ReverseRule
from comb_spec_searcher.utils import taylor_expand
from permuta import Av, Perm
from tilings import GriddedPerm, Tiling
from tilings import strategies as strat
from tilings.strategies.fusion import ComponentFusionStrategy, FusionStrategy
from tilings.strategy_pack import TileScopePack
from tilings.tilescope import GuidedSearcher, TileScope

point_placements = TileScopePack.point_placements()
all_the_strategies = TileScopePack.all_the_strategies()
all_the_strategies_fusion = TileScopePack.all_the_strategies().make_fusion(
    tracked=False
)
point_placements_fusion = point_placements.make_fusion(tracked=False)
point_placements_component_fusion = point_placements.make_fusion(
    component=True, tracked=False
)
row_placements_fusion = TileScopePack.row_and_col_placements(row_only=True).make_fusion(
    tracked=True
)
row_and_col_placements_component_fusion_fusion = (
    TileScopePack.row_and_col_placements()
    .make_fusion(component=True, tracked=False)
    .make_fusion(tracked=False)
)
reginsenc = TileScopePack.regular_insertion_encoding(3)


def collect_before(func):
    """
    Run gc collection before running the test.

    This ensure that the collection ran in the test won't take to much time.
    """

    def inner():
        gc.collect()
        func()

    return inner


@pytest.mark.timeout(20)
def test_132():
    searcher = TileScope("132", point_placements)
    spec = searcher.auto_search(smallest=True)
    assert isinstance(spec, CombinatorialSpecification)


@pytest.mark.timeout(20)
def test_132_genf():
    searcher = TileScope([Perm((0, 2, 1))], point_placements)
    spec = searcher.auto_search(smallest=True)
    spec = spec.expand_verified()
    av = Av([Perm((0, 2, 1))])
    for i in range(10):
        assert set(av.of_length(i)) == set(
            gp.patt for gp in spec.generate_objects_of_size(i)
        )
        assert spec.random_sample_object_of_size(i).patt in av
    gf = spec.get_genf()
    gf = sympy.series(spec.get_genf(), n=15)
    x = sympy.Symbol("x")
    assert [gf.coeff(x, n) for n in range(13)] == [
        1,
        1,
        2,
        5,
        14,
        42,
        132,
        429,
        1430,
        4862,
        16796,
        58786,
        208012,
    ]


@pytest.mark.timeout(20)
def test_132_elementary():
    searcher = TileScope(Tiling.from_string("132"), point_placements.make_elementary())
    spec = searcher.auto_search()
    assert spec.number_of_rules() == 4
    spec = spec.expand_verified()
    assert spec.number_of_rules() == 5
    assert isinstance(spec, CombinatorialSpecification)


@pytest.mark.timeout(20)
def test_132_321_genf():
    searcher = TileScope("132_321", point_placements)
    spec = searcher.auto_search(smallest=True)
    assert isinstance(spec, CombinatorialSpecification)
    gf = spec.get_genf()
    assert taylor_expand(gf, 15) == [
        1,
        1,
        2,
        4,
        7,
        11,
        16,
        22,
        29,
        37,
        46,
        56,
        67,
        79,
        92,
        106,
    ]


@pytest.mark.timeout(20)
def test_123():
    searcher = TileScope((Perm((0, 1, 2)),), point_placements_fusion)
    spec = searcher.auto_search(smallest=True)
    assert isinstance(spec, CombinatorialSpecification)


@pytest.mark.timeout(20)
def test_1342_1423():
    searcher = TileScope("1342_1423", point_placements_component_fusion)
    spec = searcher.auto_search(smallest=True)
    assert spec.number_of_rules() == 9
    assert isinstance(spec, CombinatorialSpecification)


@pytest.mark.timeout(60)
def test_reverse_equiv():
    """A specification that should use reverse equivalence."""
    pack = TileScopePack(
        initial_strats=[
            strat.FactorFactory(),
            strat.RequirementCorroborationFactory(),
            strat.RequirementPlacementFactory(partial=False),
        ],
        inferral_strats=[strat.RowColumnSeparationStrategy()],
        expansion_strats=[[strat.CellInsertionFactory()]],
        ver_strats=[strat.BasicVerificationStrategy()],
        iterative=False,
        name="test pack",
    )
    basis = (Perm((0, 1, 3, 2)), Perm((0, 2, 3, 1)), Perm((1, 0, 3, 2)))
    # From https://oeis.org/A033321
    expected_enum = [1, 1, 2, 6, 21, 79, 311, 1265, 5275, 22431, 96900, 424068, 1876143]
    x, f = sympy.symbols("x f")
    expected_min_poly = sympy.sympify("-4*f^2*x^2 + 8*f^2*x - 4*f*x - 4*f + 4")
    searcher = TileScope(basis, pack)
    spec = searcher.auto_search(smallest=True)
    assert [spec.count_objects_of_size(i) for i in range(13)] == expected_enum
    genf = spec.get_genf()
    assert sympy.simplify(expected_min_poly.subs(f, genf)) == 0
    assert taylor_expand(genf, 12) == expected_enum
    # In order to avoid ReccursionError we go incrementally
    for i in range(0, 100):
        spec.count_objects_of_size(i)
    assert spec.count_objects_of_size(50) == 86055297645519796258217673160170
    assert (
        spec.count_objects_of_size(100)
        == 2733073112795720153237297124938915907723365837935699807314396095313
    )
    len4_perms = tuple(spec.generate_objects_of_size(4))
    assert len(len4_perms) == 21
    assert all(p not in len4_perms for p in basis)
    len8_perms = tuple(spec.generate_objects_of_size(8))
    assert len(len8_perms) == 5275
    assert len(set(len8_perms)) == 5275
    for _ in range(10):
        gp = spec.random_sample_object_of_size(10)
        print(gp)
        assert gp.patt.avoids(*basis)

    av = Av(basis)
    for i in range(10):
        assert set(av.of_length(i)) == set(
            gp.patt for gp in spec.generate_objects_of_size(i)
        )
        assert spec.random_sample_object_of_size(i).patt in av


@pytest.mark.timeout(20)
def test_1324():
    searcher = TileScope("1324", row_and_col_placements_component_fusion_fusion)
    spec = searcher.auto_search(smallest=True)
    assert spec.number_of_rules() == 9
    num_fusion = 0
    num_comp_fusion = 0
    for rule in spec.rules_dict.values():
        if isinstance(rule.strategy, FusionStrategy) and not isinstance(
            rule.strategy, ComponentFusionStrategy
        ):
            num_fusion += 1
        if isinstance(rule.strategy, ComponentFusionStrategy):
            num_comp_fusion += 1
    assert num_fusion == 1
    assert num_comp_fusion == 1
    assert isinstance(spec, CombinatorialSpecification)


@pytest.mark.timeout(20)
def test_321_1324():
    searcher = TileScope("321_1324", reginsenc)
    spec = searcher.auto_search()
    assert isinstance(spec, CombinatorialSpecification)
    for i in range(20):
        gp = spec.random_sample_object_of_size(i)
        assert all(cell == (0, 0) for cell in gp.pos)
        assert gp.patt.avoids(Perm((2, 1, 0)), Perm((0, 2, 1, 3)))
    av = Av([Perm((2, 1, 0)), Perm((0, 2, 1, 3))])
    for i in range(10):
        assert set(av.of_length(i)) == set(
            gp.patt for gp in spec.generate_objects_of_size(i)
        )
    assert [spec.count_objects_of_size(i) for i in range(50)] == [
        1,
        1,
        2,
        5,
        13,
        32,
        72,
        148,
        281,
        499,
        838,
        1343,
        2069,
        3082,
        4460,
        6294,
        8689,
        11765,
        15658,
        20521,
        26525,
        33860,
        42736,
        53384,
        66057,
        81031,
        98606,
        119107,
        142885,
        170318,
        201812,
        237802,
        278753,
        325161,
        377554,
        436493,
        502573,
        576424,
        658712,
        750140,
        851449,
        963419,
        1086870,
        1222663,
        1371701,
        1534930,
        1713340,
        1907966,
        2119889,
        2350237,
    ]


@pytest.mark.timeout(5)
def test_from_tiling():
    t = Tiling(
        obstructions=[
            GriddedPerm((0, 1), ((0, 0), (0, 0))),
            GriddedPerm((0, 1), ((0, 0), (1, 1))),
            GriddedPerm((0, 1), ((1, 1), (1, 1))),
        ]
    )
    searcher = TileScope(t, TileScopePack.point_placements())
    spec = searcher.auto_search()
    print(spec)
    assert sympy.simplify(spec.get_genf() - sympy.sympify("(1+x)/(1-x)")) == 0


@collect_before
@pytest.mark.timeout(5)
def test_expansion():
    """
    This cannot be expanded automatically since it requires the positive
    root that is not in the specification.
    """
    pack = TileScopePack.only_root_placements(3, 1)
    css = TileScope("132", pack)
    print(pack)
    print(css)
    spec = css.auto_search(smallest=True)
    spec = spec.expand_verified()
    assert sum(1 for rule in spec if isinstance(rule, ReverseRule)) == 1
    assert [spec.count_objects_of_size(i) for i in range(13)] == [
        1,
        1,
        2,
        5,
        14,
        42,
        132,
        429,
        1430,
        4862,
        16796,
        58786,
        208012,
    ]


@pytest.mark.timeout(30)
def test_domino():
    domino = Tiling(
        obstructions=[
            GriddedPerm((0, 2, 1), [(0, 0), (0, 0), (0, 0)]),
            GriddedPerm((1, 0, 2), [(0, 1), (0, 1), (0, 1)]),
            GriddedPerm((0, 2, 1, 3), [(0, 0), (0, 1), (0, 0), (0, 1)]),
        ]
    )
    tilescope = TileScope(
        domino,
        TileScopePack.row_and_col_placements().make_fusion(
            tracked=True, component=True
        ),
    )
    spec = tilescope.auto_search()
    assert isinstance(spec, CombinatorialSpecification)
    assert [spec.count_objects_of_size(i) for i in range(15)] == [
        1,
        2,
        6,
        22,
        91,
        408,
        1938,
        9614,
        49335,
        260130,
        1402440,
        7702632,
        42975796,
        243035536,
        1390594458,
    ]


@collect_before
@pytest.mark.timeout(60)
def test_parallel_forest():
    expected_count = [1, 1, 2, 6, 22, 90, 394, 1806, 8558, 41586]
    pack = TileScopePack.only_root_placements(2, 1)
    bases = ["0231_2031", "0132_1032", "0231_0321"]
    searchers = [TileScope(b, pack, ruledb=RuleDBForest()) for b in bases]
    specs = [css.auto_search() for css in searchers]
    for spec in specs:
        spec = spec.expand_verified()
        count = [spec.count_objects_of_size(n) for n in range(10)]
        assert count == expected_count


@collect_before
@pytest.mark.timeout(15)
def forest_expansion():
    """
    A forest spec that at some point became not productive once expanded.
    """
    basis = "0213_1032_1302"
    pack = TileScopePack.point_placements()
    css = TileScope(basis, pack, ruledb=RuleDBForest())
    spec = css.auto_search(status_update=30)
    for rule in spec.rules_dict.values():
        rule.label = spec.get_label(rule.comb_class)
    spec.expand_verified()
    assert [spec.count_objects_of_size(i) for i in range(13)] == [
        1,
        1,
        2,
        6,
        21,
        77,
        287,
        1079,
        4082,
        15522,
        59280,
        227240,
        873886,
    ]


def test_guided_searcher():
    tilescope = TileScope(
        "123", TileScopePack.point_placements().make_fusion(tracked=False)
    )
    spec = tilescope.auto_search()
    searcher = GuidedSearcher.from_spec(
        spec, TileScopePack.point_placements().make_fusion(tracked=True)
    )
    tracked_spec = searcher.auto_search()
    tracked_spec = tracked_spec.expand_verified()
    assert [tracked_spec.count_objects_of_size(i) for i in range(13)] == [
        1,
        1,
        2,
        5,
        14,
        42,
        132,
        429,
        1430,
        4862,
        16796,
        58786,
        208012,
    ]



from typing import Iterable, List, Optional, Tuple, Type, TypeVar

from comb_spec_searcher import VerificationStrategy
from permuta import Perm
from permuta.permutils.symmetry import all_symmetry_sets
from tilings import GriddedPerm, Tiling

__all__ = [
    "BasisAwareVerificationStrategy",
]


BasisAwareVerificationStrategyType = TypeVar(
    "BasisAwareVerificationStrategyType", bound="BasisAwareVerificationStrategy"
)


class BasisAwareVerificationStrategy(VerificationStrategy[Tiling, GriddedPerm]):
    """
    A base class for a verification strategy that needs to know the basis the
    Tilescope is currently running.
    """

    def __init__(
        self,
        basis: Optional[Iterable[Perm]] = None,
        symmetry: bool = False,
        ignore_parent: bool = False,
    ):
        self._basis = tuple(basis) if basis is not None else tuple()
        self._symmetry = symmetry
        assert all(
            isinstance(p, Perm) for p in self._basis
        ), "Element of the basis must be Perm"
        if symmetry:
            self.symmetries = frozenset(
                frozenset(b) for b in all_symmetry_sets(self._basis)
            )
        else:
            self.symmetries = frozenset([frozenset(self._basis)])
        super().__init__(ignore_parent=ignore_parent)

    def change_basis(
        self: BasisAwareVerificationStrategyType, basis: Iterable[Perm], symmetry: bool
    ) -> BasisAwareVerificationStrategyType:
        """
        Return a new version of the verification strategy with the given basis instead
        of the current one.
        """
        basis = tuple(basis)
        return self.__class__(basis, symmetry, self.ignore_parent)

    @property
    def basis(self) -> Tuple[Perm, ...]:
        return self._basis

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["basis"] = self._basis
        d["symmetry"] = self._symmetry
        return d

    @classmethod
    def from_dict(
        cls: Type[BasisAwareVerificationStrategyType], d: dict
    ) -> BasisAwareVerificationStrategyType:
        if "basis" in d and d["basis"] is not None:
            basis: Optional[List[Perm]] = [Perm(p) for p in d.pop("basis")]
        else:
            basis = d.pop("basis", None)
        return cls(basis=basis, **d)

    def __repr__(self) -> str:
        args = ", ".join(
            [
                f"basis={self._basis}",
                f"symmetry={self._symmetry}",
                f"ignore_parent={self.ignore_parent}",
            ]
        )
        return f"{self.__class__.__name__}({args})"



from collections import Counter
from functools import partial
from itertools import combinations
from typing import Callable, Dict, Iterator, List, Optional, Tuple, Union

import sympy

from comb_spec_searcher import (
    Constructor,
    DisjointUnionStrategy,
    Strategy,
    StrategyFactory,
)
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.typing import (
    Parameters,
    ParametersMap,
    RelianceProfile,
    SubObjects,
    SubRecs,
    SubSamplers,
    SubTerms,
    Terms,
)
from tilings import GriddedPerm, Tiling
from tilings.assumptions import (
    ComponentAssumption,
    SkewComponentAssumption,
    SumComponentAssumption,
    TrackingAssumption,
)

Cell = Tuple[int, int]


class RearrangeConstructor(Constructor[Tiling, GriddedPerm]):
    def __init__(
        self,
        parent: Tiling,
        child: Tiling,
        assumption: TrackingAssumption,
        sub_assumption: TrackingAssumption,
        extra_parameters: Dict[str, str],
    ):
        """
        Constructor for the rearrange strategy.

        The extra_parameters should a dict mapping variable on the parent to the
        associated variable on the child. The variable for `assumption` should not
        appear in the dict since it does not match directly to a variable on the child.
        """
        new_ass = TrackingAssumption(set(assumption.gps) - set(sub_assumption.gps))
        self.new_ass_child_idx = child.extra_parameters.index(
            child.get_assumption_parameter(new_ass)
        )
        self.ass_parent_idx = parent.extra_parameters.index(
            parent.get_assumption_parameter(assumption)
        )
        self.subass_parent_idx = parent.extra_parameters.index(
            parent.get_assumption_parameter(sub_assumption)
        )
        self.subass_child_idx = child.extra_parameters.index(
            child.get_assumption_parameter(sub_assumption)
        )
        self.child_to_parent_param_map = self._build_child_to_parent_param_map(
            parent,
            child,
            extra_parameters,
            assumption,
            sub_assumption,
        )
        self.parent_to_child_param_map = self._build_parent_to_child_param_map(
            parent,
            child,
            extra_parameters,
            assumption,
            sub_assumption,
        )
        self.parent_dict_to_param = self._build_map_dict_to_param(parent)
        self.child_param_to_dict = self._build_map_param_to_dict(child)
        self.eq_subs = self._build_eq_subs(parent, child, extra_parameters)

    def _build_child_to_parent_param_map(
        self,
        parent: Tiling,
        child: Tiling,
        extra_parameters: Dict[str, str],
        assumptions: TrackingAssumption,
        sub_assumption: TrackingAssumption,
    ) -> ParametersMap:
        """
        Build a maps that maps parameters on the child to the corresponding parameters
        on the parent.
        """
        reversed_extra_param = {v: k for k, v in extra_parameters.items()}
        parent_param_to_pos = {
            param: pos for pos, param in enumerate(parent.extra_parameters)
        }
        child_pos_to_parent_pos: List[Tuple[int, ...]] = []
        for pos, param in enumerate(child.extra_parameters):
            to_add: List[int] = []
            if pos == self.subass_child_idx:
                to_add.append(self.ass_parent_idx)
            elif pos == self.new_ass_child_idx:
                to_add.append(self.ass_parent_idx)
            if param in reversed_extra_param:
                to_add.append(parent_param_to_pos[reversed_extra_param[param]])
            child_pos_to_parent_pos.append(tuple(to_add))
        return self.build_param_map(
            tuple(child_pos_to_parent_pos), len(parent.extra_parameters)
        )

    def param_map_for_rearrange(
        self,
        parent_pos_to_child_pos: Tuple[Tuple[int, int], ...],
        num_child_param: int,
        param: Parameters,
    ) -> Parameters:
        new_param = [-1 for _ in range(num_child_param)]
        for ppos, cpos in parent_pos_to_child_pos:
            new_param[cpos] = param[ppos]
        new_ass_value = param[self.ass_parent_idx] - param[self.subass_parent_idx]
        assert new_param[self.new_ass_child_idx] in (-1, new_ass_value)
        new_param[self.new_ass_child_idx] = new_ass_value
        assert all(v >= 0 for v in new_param)
        return tuple(new_param)

    def _build_parent_to_child_param_map(
        self,
        parent: Tiling,
        child: Tiling,
        extra_parameters: Dict[str, str],
        assumptions: TrackingAssumption,
        sub_assumption: TrackingAssumption,
    ) -> ParametersMap:
        """
        Build a maps that maps parameters on the parent to the corresponding parameters
        on the child.
        """
        num_child_param = len(child.extra_parameters)
        # Each pair in the tuple indicate the parent param -> child param
        parent_pos_to_child_pos: Tuple[Tuple[int, int], ...] = tuple(
            (
                parent.extra_parameters.index(pparam),
                child.extra_parameters.index(cparam),
            )
            for pparam, cparam in extra_parameters.items()
        )

        return partial(
            self.param_map_for_rearrange, parent_pos_to_child_pos, num_child_param
        )

    @staticmethod
    def _dict_to_param_helper(tiling_params_order, d):
        """
        this is required to be a standalone function to be pickleable
        """
        return tuple(d[p] for p in tiling_params_order)

    @staticmethod
    def _build_map_dict_to_param(
        tiling: Tiling,
    ) -> Callable[[Dict[str, int]], Parameters]:
        """
        Returns a map the return the param tuple  from the param dictionary for
        the given tilings.
        Should probably live somewhere else.
        """

        tiling_params_order = tiling.extra_parameters
        return partial(RearrangeConstructor._dict_to_param_helper, tiling_params_order)

    @staticmethod
    def _param_to_dict_helper(tiling_params_order, p):
        """
        this is required to be a standalone function to be pickleable
        """
        return dict(zip(tiling_params_order, p))

    @staticmethod
    def _build_map_param_to_dict(
        tiling: Tiling,
    ) -> Callable[[Parameters], Dict[str, int]]:
        """
        Returns a map the return the param dictionary  from the param tuple for the
        given tilings.
        Should probably live somewhere else.
        """
        tiling_params_order = tiling.extra_parameters
        return partial(RearrangeConstructor._param_to_dict_helper, tiling_params_order)

    def _build_eq_subs(
        self, parent: Tiling, child: Tiling, extra_parameters: Dict[str, str]
    ) -> dict:
        """
        Build the substitution dict needed to compute the equation.
        """
        subs = {
            sympy.var(child): sympy.var(parent)
            for parent, child in extra_parameters.items()
        }
        new_ass_var_child = sympy.var(child.extra_parameters[self.new_ass_child_idx])
        ass_var_parent = sympy.var(parent.extra_parameters[self.ass_parent_idx])
        subass_var_child = sympy.var(child.extra_parameters[self.subass_child_idx])
        subs[new_ass_var_child] = subs.get(new_ass_var_child, 1) * ass_var_parent
        subs[subass_var_child] *= ass_var_parent
        return subs

    def get_equation(
        self, lhs_func: sympy.Function, rhs_funcs: Tuple[sympy.Function, ...]
    ) -> sympy.Eq:
        rhs = rhs_funcs[0].subs(self.eq_subs, simultaneous=True)
        return sympy.Eq(lhs_func, rhs)

    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        raise NotImplementedError

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        terms: Terms = Counter()
        for param, value in subterms[0](n).items():
            new_param = self.child_to_parent_param_map(param)
            terms[new_param] += value
        return terms

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[Tuple[Parameters, Tuple[List[Optional[GriddedPerm]]]]]:
        assert len(subobjs) == 1
        for param, objs in subobjs[0](n).items():
            new_param = self.child_to_parent_param_map(param)
            for obj in objs:
                yield new_param, ([obj],)

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ) -> Tuple[GriddedPerm]:
        assert len(subsamplers) == 1
        parent_param_tuple = self.parent_dict_to_param(parameters)
        child_param_tuple = self.parent_to_child_param_map(parent_param_tuple)
        child_param_dict = self.child_param_to_dict(child_param_tuple)
        return (subsamplers[0](n, **child_param_dict),)

    def equiv(
        self, other: "Constructor", data: Optional[object] = None
    ) -> Tuple[bool, Optional[object]]:
        return isinstance(other, type(self)), None


class ReverseRearrangeConstructor(RearrangeConstructor):
    def __init__(
        self,
        parent: Tiling,
        child: Tiling,
        assumption: TrackingAssumption,
        sub_assumption: TrackingAssumption,
        extra_parameters: Dict[str, str],
    ):
        super().__init__(parent, child, assumption, sub_assumption, extra_parameters)
        self.child_to_parent_param_map, self.parent_to_child_param_map = (
            self.parent_to_child_param_map,
            self.child_to_parent_param_map,
        )
        self.parent_dict_to_param = self._build_map_dict_to_param(child)
        self.child_param_to_dict = self._build_map_param_to_dict(parent)

    def get_equation(
        self, lhs_func: sympy.Function, rhs_funcs: Tuple[sympy.Function, ...]
    ) -> sympy.Eq:
        return super().get_equation(rhs_funcs[0], (lhs_func,))


class RearrangeAssumptionStrategy(Strategy[Tiling, GriddedPerm]):
    def __init__(
        self, assumption: TrackingAssumption, sub_assumption: TrackingAssumption
    ):
        self.assumption = assumption
        self.sub_assumption = sub_assumption
        super().__init__()

    def can_be_equivalent(self) -> bool:
        return False

    def is_two_way(self, comb_class: Tiling) -> bool:
        return True

    def is_reversible(self, comb_class: Tiling) -> bool:
        return True

    def shifts(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[int, ...]:
        return (0,)

    def reverse_constructor(
        self,
        idx: int,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> ReverseRearrangeConstructor:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Can't split the tracking assumption")
        return ReverseRearrangeConstructor(
            comb_class,
            children[0],
            self.assumption,
            self.sub_assumption,
            self.extra_parameters(comb_class, children)[0],
        )

    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling]:
        tiling = comb_class.remove_assumption(self.assumption)
        new_ass1 = TrackingAssumption(
            set(self.assumption.gps) - set(self.sub_assumption.gps)
        )
        return (tiling.add_assumption(new_ass1),)

    def constructor(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> RearrangeConstructor:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Can't split the tracking assumption")
        return RearrangeConstructor(
            comb_class,
            children[0],
            self.assumption,
            self.sub_assumption,
            self.extra_parameters(comb_class, children)[0],
        )

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str]]:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        res: Dict[str, str] = {}
        child = children[0]
        for parent_ass, parent_param in zip(
            comb_class.assumptions, comb_class.extra_parameters
        ):
            if parent_ass == self.assumption:
                continue
            child_param = child.extra_parameters[child.assumptions.index(parent_ass)]
            res[parent_param] = child_param
        return (res,)

    def formal_step(self) -> str:
        return f"rearranging the assumption {self.assumption} and {self.sub_assumption}"

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        """
        The backward direction of the underlying bijection used for object
        generation and sampling.
        """
        assert len(objs) == 1 and objs[0] is not None
        yield objs[0]

    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm], ...]:
        """
        The forward direction of the underlying bijection used for object
        generation and sampling.
        """
        return (obj,)

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d.pop("ignore_parent")
        d.pop("inferrable")
        d.pop("possibly_empty")
        d.pop("workable")
        d["assumption"] = self.assumption.to_jsonable()
        d["sub_assumption"] = self.sub_assumption.to_jsonable()
        return d

    def __repr__(self) -> str:
        args = ", ".join(
            [
                f"assumption={self.assumption!r}",
                f"sub_assumption={self.sub_assumption!r}",
            ]
        )
        return f"{self.__class__.__name__}({args})"

    @classmethod
    def from_dict(cls, d: dict) -> "RearrangeAssumptionStrategy":
        assumption = TrackingAssumption.from_dict(d.pop("assumption"))
        sub_assumption = TrackingAssumption.from_dict(d.pop("sub_assumption"))
        assert not d
        return cls(assumption, sub_assumption)

    @staticmethod
    def get_eq_symbol() -> str:
        return "↣"


class ComponentToPointAssumptionStrategy(
    DisjointUnionStrategy[Tiling, GriddedPerm],
):
    """A strategy that changes a component tracking assumption to a point
    tracking assumption."""

    def __init__(
        self,
        assumption: TrackingAssumption,
        ignore_parent: bool = False,
        workable: bool = False,
    ):
        assert isinstance(assumption, ComponentAssumption)
        self.assumption = assumption
        self.new_assumption = TrackingAssumption(assumption.gps)
        super().__init__(
            ignore_parent=ignore_parent,
            inferrable=False,
            possibly_empty=False,
            workable=workable,
        )

    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling]:
        if self.assumption not in comb_class.assumptions:
            raise StrategyDoesNotApply("Assumption not on tiling")
        return (
            comb_class.remove_assumption(self.assumption).add_assumption(
                self.new_assumption
            ),
        )

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        assert len(objs) == 1 and objs[0] is not None
        yield objs[0]

    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm]]:
        return (obj,)

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str], ...]:
        if not comb_class.extra_parameters:
            return super().extra_parameters(comb_class, children)
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        child = children[0]
        return (
            {
                comb_class.get_assumption_parameter(
                    ass
                ): child.get_assumption_parameter(
                    self.new_assumption if ass == self.assumption else ass
                )
                for ass in comb_class.assumptions
            },
        )

    def formal_step(self) -> str:
        cells = ", ".join(str(gp.pos[0]) for gp in self.assumption.gps)
        return f"change component assumption in cells {cells} to point assumption"

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["assumption"] = self.assumption.to_jsonable()
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "ComponentToPointAssumptionStrategy":
        return cls(
            assumption=TrackingAssumption.from_dict(d["assumption"]),
            ignore_parent=d["ignore_parent"],
            workable=d["workable"],
        )


class RearrangeAssumptionFactory(StrategyFactory[Tiling]):
    def __call__(
        self, comb_class: Tiling
    ) -> Iterator[
        Union[RearrangeAssumptionStrategy, ComponentToPointAssumptionStrategy]
    ]:
        points: List[TrackingAssumption] = []
        components: List[TrackingAssumption] = []
        for ass in comb_class.assumptions:
            (points, components)[isinstance(ass, ComponentAssumption)].append(ass)

        for ass1, ass2 in combinations(points, 2):
            if set(ass1.gps).issubset(set(ass2.gps)):
                yield RearrangeAssumptionStrategy(ass2, ass1)
            if set(ass2.gps).issubset(set(ass1.gps)):
                yield RearrangeAssumptionStrategy(ass1, ass2)

        for ass in components:
            if self.can_be_point_assumption(comb_class, ass):
                yield ComponentToPointAssumptionStrategy(ass)

    @staticmethod
    def can_be_point_assumption(tiling: Tiling, assumption: TrackingAssumption) -> bool:
        sub_tiling = tiling.sub_tiling(tuple(gp.pos[0] for gp in assumption.gps))
        if isinstance(assumption, SumComponentAssumption):
            return sub_tiling.is_increasing()
        assert isinstance(assumption, SkewComponentAssumption)
        return sub_tiling.is_decreasing()

    def __repr__(self) -> str:
        return self.__class__.__name__ + "()"

    def __str__(self) -> str:
        return "rearrange assumptions"

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "RearrangeAssumptionFactory":
        return cls()



from typing import Dict, Iterator, Optional, Tuple

from comb_spec_searcher import Strategy
from tilings import GriddedPerm, Tiling

from .dummy_constructor import DummyConstructor


class DummyStrategy(Strategy[Tiling, GriddedPerm]):
    def can_be_equivalent(self) -> bool:
        raise NotImplementedError

    def is_two_way(self, comb_class: Tiling) -> bool:
        raise NotImplementedError

    def is_reversible(self, comb_class: Tiling) -> bool:
        raise NotImplementedError

    def shifts(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]],
    ) -> Tuple[int, ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
        raise NotImplementedError

    def decomposition_function(
        self, comb_class: Tiling
    ) -> Optional[Tuple[Tiling, ...]]:
        raise NotImplementedError

    def formal_step(self) -> str:
        return "dummy strategy"

    def constructor(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> DummyConstructor:
        if children is None:
            children = self.decomposition_function(comb_class)
        return DummyConstructor()

    def reverse_constructor(
        self,
        idx: int,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> DummyConstructor:
        if children is None:
            children = self.decomposition_function(comb_class)
        return DummyConstructor()

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        if children is None:
            children = self.decomposition_function(comb_class)
        raise NotImplementedError

    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm], ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
        raise NotImplementedError

    def extra_parameters(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Dict[str, str], ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
        raise NotImplementedError

    @classmethod
    def from_dict(cls, d: dict) -> "DummyStrategy":
        return cls(**d)



"""
Collection of function that are not directly related to the code but still
useful.
"""
from functools import reduce
from typing import (
    Collection,
    Dict,
    Iterable,
    Iterator,
    List,
    Sequence,
    Set,
    Tuple,
    TypeVar,
)

Vertex = TypeVar("Vertex")
T = TypeVar("T")
AdjTable = Dict[Vertex, Set[Vertex]]
Cell = Tuple[int, int]


def union_reduce(iterables: Iterable[Iterable[T]]) -> Set[T]:
    """
    Returns the union of the elements contained in the iterables.

    >>> sorted(union_reduce(([1,2], [2,3])))
    [1, 2, 3]
    >>> sorted(union_reduce(([], range(0, 10, 2))))
    [0, 2, 4, 6, 8]
    >>> union_reduce(([], []))
    set()
    """
    sets: Iterator[Set[T]] = map(set, iterables)
    return reduce(set.__or__, sets, set())


def intersection_reduce(iterables: Iterable[Iterable[T]]) -> Set[T]:
    """Returns the intersection of the iterables."""
    try:
        sets: Iterator[Set[T]] = map(set, iterables)
        return reduce(set.__and__, sets)
    except TypeError:
        return set()


def is_tree(
    vertices: Collection[Vertex], edges: Collection[Tuple[Vertex, Vertex]]
) -> bool:
    """
    Return True if the undirected graph is a tree.

    That is, it is has no cycles and is connected.
    """
    if not vertices:
        return True
    adj_table = adjacency_table(vertices, edges)
    return len(edges) + 1 == len(vertices) and is_connected(adj_table)


def adjacency_table(
    vertices: Collection[Vertex], edges: Collection[Tuple[Vertex, Vertex]]
) -> AdjTable:
    """Return adjacency table of edges."""
    adj_table = {v: set() for v in vertices}  # type: AdjTable
    for c1, c2 in edges:
        adj_table[c1].add(c2)
        adj_table[c2].add(c1)
    return adj_table


def is_connected(adj_table: AdjTable) -> bool:
    """Return True if graph with adjacency table is connected."""
    if not adj_table:
        return True
    visited = {cell: False for cell in adj_table}
    start_vertex = next(iter(adj_table.keys()))
    visited[start_vertex] = True
    stack = [start_vertex]
    while stack:
        curr = stack.pop()
        for vertex in adj_table[curr]:
            if not visited[vertex]:
                stack.append(vertex)
                visited[vertex] = True
    return all(visited.values())


def multinomial(lst):
    """
    Returns the multinomial. Taken from
    https://stackoverflow.com/questions/46374185/does-python-have-a-function-which-computes-multinomial-coefficients
    """
    res, i = 1, sum(lst)
    i0 = lst.index(max(lst))
    for a in lst[:i0] + lst[i0 + 1 :]:
        for j in range(1, a + 1):
            res *= i
            res //= j
            i -= 1
    return res


# The code below is magical and comes from
# https://codereview.stackexchange.com/questions/1526/finding-all-k-subset-partitions


def partitions_iterator(lst: Sequence[T]) -> Iterator[Tuple[Tuple[T, ...], ...]]:
    """
    Iterator over all the possible partitions of a list. A partition is yielded
    as a list of list.

    The partition in a single part and the parition consisting of only
    singleton are not returned.

    >>> for partition in partitions_iterator([1, 2, 3]):
    ...     print(partition)
    ((1, 2), (3,))
    ((1,), (2, 3))
    ((1, 3), (2,))
    """
    for i in range(2, len(lst)):
        for part in algorithm_u(lst, i):
            yield tuple(map(tuple, part))


def algorithm_u(ns: Sequence[T], m: int):
    # pylint: disable=too-many-statements,too-many-branches
    def visit(n, a):
        ps: List[List[T]] = [[] for i in range(m)]
        for j in range(n):
            ps[a[j + 1]].append(ns[j])
        return ps

    def f(mu, nu, sigma, n, a):
        if mu == 2:
            yield visit(n, a)
        else:
            yield from f(mu - 1, nu - 1, (mu + sigma) % 2, n, a)
        if nu == mu + 1:
            a[mu] = mu - 1
            yield visit(n, a)
            while a[nu] > 0:
                a[nu] = a[nu] - 1
                yield visit(n, a)
        elif nu > mu + 1:
            if (mu + sigma) % 2 == 1:
                a[nu - 1] = mu - 1
            else:
                a[mu] = mu - 1
            if (a[nu] + sigma) % 2 == 1:
                yield from b(mu, nu - 1, 0, n, a)
            else:
                yield from f(mu, nu - 1, 0, n, a)
            while a[nu] > 0:
                a[nu] = a[nu] - 1
                if (a[nu] + sigma) % 2 == 1:
                    yield from b(mu, nu - 1, 0, n, a)
                else:
                    yield from f(mu, nu - 1, 0, n, a)

    def b(mu, nu, sigma, n, a):
        if nu == mu + 1:
            while a[nu] < mu - 1:
                yield visit(n, a)
                a[nu] = a[nu] + 1
            yield visit(n, a)
            a[mu] = 0
        elif nu > mu + 1:
            if (a[nu] + sigma) % 2 == 1:
                yield from f(mu, nu - 1, 0, n, a)
            else:
                yield from b(mu, nu - 1, 0, n, a)
            while a[nu] < mu - 1:
                a[nu] = a[nu] + 1
                if (a[nu] + sigma) % 2 == 1:
                    yield from f(mu, nu - 1, 0, n, a)
                else:
                    yield from b(mu, nu - 1, 0, n, a)
            if (mu + sigma) % 2 == 1:
                a[nu - 1] = 0
            else:
                a[mu] = 0
        if mu == 2:
            yield visit(n, a)
        else:
            yield from b(mu - 1, nu - 1, (mu + sigma) % 2, n, a)

    n = len(ns)
    a = [0] * (n + 1)
    for j in range(1, m + 1):
        a[n - m + j] = j - 1
    return f(m, n, 0, n, a)



from collections import Counter
from itertools import product
from random import randint
from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Tuple

from sympy import Eq, Expr, Function, Number, Symbol, var

from comb_spec_searcher import Constructor, Strategy, StrategyFactory
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.strategies import Rule
from comb_spec_searcher.typing import (
    Parameters,
    ParametersMap,
    RelianceProfile,
    SubObjects,
    SubRecs,
    SubSamplers,
    SubTerms,
    Terms,
)
from tilings import GriddedPerm, Tiling
from tilings.assumptions import ComponentAssumption, TrackingAssumption

Cell = Tuple[int, int]


class AddAssumptionsConstructor(Constructor):
    """
    The constructor used to count when a new variable is added.
    """

    def __init__(
        self,
        parent: Tiling,
        child: Tiling,
        new_parameters: Iterable[str],
        extra_parameters: Dict[str, str],
    ):
        #  parent parameter -> child parameter mapping
        self.extra_parameters = extra_parameters
        #  the paramater that was added, to count we must sum over all possible values
        self.new_parameters = tuple(new_parameters)
        self.child_param_map = self._build_child_param_map(parent, child)

    def get_equation(self, lhs_func: Function, rhs_funcs: Tuple[Function, ...]) -> Eq:
        rhs_func = rhs_funcs[0]
        subs: Dict[Any, Expr] = {
            var(child): var(parent) for parent, child in self.extra_parameters.items()
        }
        for k in self.new_parameters:
            subs[k] = Number(1)
        return Eq(lhs_func, rhs_func.subs(subs, simultaneous=True))

    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        raise NotImplementedError

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        assert len(subterms) == 1
        return self._push_add_assumption(n, subterms[0], self.child_param_map)

    @staticmethod
    def _push_add_assumption(
        n: int,
        child_terms: Callable[[int], Terms],
        child_param_map: ParametersMap,
    ) -> Terms:
        new_terms: Terms = Counter()
        for param, value in child_terms(n).items():
            new_terms[child_param_map(param)] += value
        return new_terms

    def _build_child_param_map(self, parent: Tiling, child: Tiling) -> ParametersMap:
        parent_param_to_pos = {
            param: pos for pos, param in enumerate(parent.extra_parameters)
        }
        child_param_to_parent_param = {v: k for k, v in self.extra_parameters.items()}
        child_pos_to_parent_pos: Tuple[Tuple[int, ...], ...] = tuple(
            tuple()
            if param in self.new_parameters
            else (parent_param_to_pos[child_param_to_parent_param[param]],)
            for param in child.extra_parameters
        )
        return self.build_param_map(
            child_pos_to_parent_pos, len(parent.extra_parameters)
        )

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[Tuple[Parameters, Tuple[List[Optional[GriddedPerm]], ...]]]:
        for param, gps in subobjs[0](n).items():
            yield self.child_param_map(param), (gps,)

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ):
        subrec = subrecs[0]
        subsampler = subsamplers[0]
        random_choice = randint(1, parent_count)
        new_params = {self.extra_parameters[k]: val for k, val in parameters.items()}
        res = 0
        for values in product(*[range(n + 1) for _ in self.new_parameters]):
            for k, val in zip(self.new_parameters, values):
                new_params[k] = val
            res += subrec(n, **new_params)
            if random_choice <= res:
                return (subsampler(n, **new_params),)

    def equiv(
        self, other: "Constructor", data: Optional[object] = None
    ) -> Tuple[bool, Optional[object]]:
        return (
            isinstance(other, type(self))
            and len(other.new_parameters) == len(self.new_parameters)
            and AddAssumptionsConstructor.extra_params_equiv(
                (self.extra_parameters,), (other.extra_parameters,)
            ),
            None,
        )


class RemoveAssumptionsConstructor(Constructor):
    """
    The constructor used to count when a variable the same as n is removed.
    """

    def __init__(
        self,
        parent: Tiling,
        child: Tiling,
        parameter: str,
        extra_parameters: Dict[str, str],
    ):
        #  parent parameter -> child parameter mapping
        self.extra_parameters = extra_parameters
        #  the paramater that was added, to count we must sum over all possible values
        self.parameter = parameter
        self.parameter_idx = parent.extra_parameters.index(self.parameter)
        self.child_param_map = self._build_child_param_map(parent, child)

    def get_equation(self, lhs_func: Function, rhs_funcs: Tuple[Function, ...]) -> Eq:
        rhs_func = rhs_funcs[0]
        subs: Dict[Symbol, Expr] = {
            var(child): var(parent) for parent, child in self.extra_parameters.items()
        }
        subs[var("x")] = var("x") * var(self.parameter)
        return Eq(lhs_func, rhs_func.subs(subs, simultaneous=True))

    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        raise NotImplementedError

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        assert len(subterms) == 1
        return self._push_add_assumption(n, subterms[0], self.child_param_map)

    def _push_add_assumption(
        self,
        n: int,
        child_terms: Callable[[int], Terms],
        child_param_map: ParametersMap,
    ) -> Terms:
        new_terms: Terms = Counter()
        for param, value in child_terms(n).items():
            new_param = list(child_param_map(param))
            new_param[self.parameter_idx] = n
            new_terms[tuple(new_param)] += value
        return new_terms

    def _build_child_param_map(self, parent: Tiling, child: Tiling) -> ParametersMap:
        parent_param_to_pos = {
            param: pos for pos, param in enumerate(parent.extra_parameters)
        }
        child_param_to_parent_param = {v: k for k, v in self.extra_parameters.items()}
        child_pos_to_parent_pos: Tuple[Tuple[int, ...], ...] = tuple(
            (parent_param_to_pos[child_param_to_parent_param[param]],)
            for param in child.extra_parameters
        )
        return self.build_param_map(
            child_pos_to_parent_pos, len(parent.extra_parameters)
        )

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[Tuple[Parameters, Tuple[List[Optional[GriddedPerm]], ...]]]:
        raise NotImplementedError

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ):
        raise NotImplementedError

    def equiv(
        self, other: "Constructor", data: Optional[object] = None
    ) -> Tuple[bool, Optional[object]]:
        return (
            isinstance(other, type(self))
            and len(other.parameter) == len(self.parameter)
            and AddAssumptionsConstructor.extra_params_equiv(
                (self.extra_parameters,), (other.extra_parameters,)
            ),
            None,
        )


class AddAssumptionsStrategy(Strategy[Tiling, GriddedPerm]):
    def __init__(self, assumptions: Iterable[TrackingAssumption], workable=False):
        self.assumptions = tuple(set(assumptions))
        super().__init__(
            ignore_parent=False,
            inferrable=True,
            possibly_empty=False,
            workable=workable,
        )

    def can_be_equivalent(self) -> bool:
        return False

    def is_two_way(self, comb_class: Tiling):
        return False

    def is_reversible(self, comb_class: Tiling) -> bool:
        return all(
            not isinstance(assumption, ComponentAssumption)
            and frozenset(gp.pos[0] for gp in assumption.gps) == comb_class.active_cells
            for assumption in self.assumptions
        )

    def shifts(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[int, ...]:
        return (0,)

    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling]:
        if any(assumption in comb_class.assumptions for assumption in self.assumptions):
            raise StrategyDoesNotApply("The assumption is already on the tiling.")
        return (comb_class.add_assumptions(self.assumptions),)

    def constructor(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> AddAssumptionsConstructor:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Can't split the tracking assumption")
        new_parameters = [
            children[0].get_assumption_parameter(ass) for ass in self.assumptions
        ]
        return AddAssumptionsConstructor(
            comb_class,
            children[0],
            new_parameters,
            self.extra_parameters(comb_class, children)[0],
        )

    def reverse_constructor(
        self,
        idx: int,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Constructor:
        if children is None:
            children = self.decomposition_function(comb_class)
        assert idx == 0
        child = children[idx]
        assert len(self.assumptions) == 1
        assumption = self.assumptions[0]
        assert len(assumption.gps) == len(comb_class.active_cells)
        parameter = child.get_assumption_parameter(self.assumptions[0])
        extra_params = {
            child.get_assumption_parameter(ass): comb_class.get_assumption_parameter(
                ass
            )
            for ass in child.assumptions
            if ass != assumption
        }

        return RemoveAssumptionsConstructor(child, comb_class, parameter, extra_params)

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str]]:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        child = children[0]
        return (
            {
                comb_class.get_assumption_parameter(
                    ass
                ): child.get_assumption_parameter(ass)
                for ass in comb_class.assumptions
            },
        )

    def formal_step(self) -> str:
        if len(self.assumptions) == 1:
            return f"adding the assumption '{self.assumptions[0]}'"
        assumptions = ", ".join([f"'{ass}'" for ass in self.assumptions])
        return f"adding the assumptions '{assumptions}'"

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        """
        The backward direction of the underlying bijection used for object
        generation and sampling.
        """
        if children is None:
            children = self.decomposition_function(comb_class)
        assert len(objs) == 1 and objs[0] is not None
        yield objs[0]

    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm], ...]:
        """
        The forward direction of the underlying bijection used for object
        generation and sampling.
        """
        if children is None:
            children = self.decomposition_function(comb_class)
        return (obj,)

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d.pop("ignore_parent")
        d.pop("inferrable")
        d.pop("possibly_empty")
        d["assumptions"] = [ass.to_jsonable() for ass in self.assumptions]
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "AddAssumptionsStrategy":
        assumptions = [TrackingAssumption.from_dict(ass) for ass in d["assumptions"]]
        return cls(assumptions)

    @staticmethod
    def get_eq_symbol() -> str:
        return "↣"

    def __repr__(self):
        return (
            self.__class__.__name__
            + f"(assumptions={repr(self.assumptions)}, workable={self.workable})"
        )


class AddAssumptionFactory(StrategyFactory[Tiling]):
    def __call__(self, comb_class: Tiling) -> Iterator[Rule]:
        for assumption in comb_class.assumptions:
            without = comb_class.remove_assumption(assumption)
            strategy = AddAssumptionsStrategy((assumption,))
            yield strategy(without)

    def __repr__(self) -> str:
        return self.__class__.__name__ + "()"

    def __str__(self) -> str:
        return "add assumptions"

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "AddAssumptionFactory":
        return cls()



import itertools
import math
from array import array
from collections import Counter
from typing import Counter as CounterType
from typing import (
    Deque,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)

import requests
import tabulate
from logzero import logger

from comb_spec_searcher import (
    CombinatorialSpecification,
    CombinatorialSpecificationSearcher,
)
from comb_spec_searcher.class_db import ClassDB, ClassKey, Info, Key
from comb_spec_searcher.class_queue import CSSQueue, DefaultQueue, WorkPacket
from comb_spec_searcher.rule_db.abstract import RuleDBAbstract
from comb_spec_searcher.strategies.rule import AbstractRule
from comb_spec_searcher.typing import CombinatorialClassType, CSSstrategy
from permuta import Basis, Perm
from tilings import GriddedPerm, Tiling
from tilings.assumptions import TrackingAssumption
from tilings.strategy_pack import TileScopePack

__all__ = ("TileScope", "TileScopePack", "LimitedAssumptionTileScope", "GuidedSearcher")

Cell = Tuple[int, int]
TrackedClassAssumption = Tuple[int, Tuple[Cell, ...]]
TrackedClassDBKey = Tuple[int, Tuple[TrackedClassAssumption, ...]]


class TileScope(CombinatorialSpecificationSearcher):
    """
    An instance of TileScope is used to build up knowledge about tilings with
    respect to the given basis.
    """

    def __init__(
        self,
        start_class: Union[str, Iterable[Perm], Tiling],
        strategy_pack: TileScopePack,
        ruledb: Optional[RuleDBAbstract] = None,
        classdb: Optional[ClassDB] = None,
        classqueue: Optional[CSSQueue] = None,
        expand_verified: bool = False,
        debug: bool = False,
    ) -> None:

        """Initialise TileScope."""
        if isinstance(start_class, Tiling):
            start_tiling = start_class
            if start_tiling.dimensions == (1, 1):
                basis = Basis(*[o.patt for o in start_tiling.obstructions])
        else:
            # Handle the non-Tiling cases
            if isinstance(start_class, str):
                basis = Basis.from_string(start_class)
            else:
                try:
                    basis = Basis(*start_class)
                except TypeError as e:
                    raise ValueError(
                        "start class must be a string, an iterable of Perm or a tiling"
                    ) from e

            # Now create start_tiling from basis
            start_tiling = Tiling(
                obstructions=[GriddedPerm.single_cell(patt, (0, 0)) for patt in basis]
            )

        if start_tiling.dimensions == (1, 1):
            logger.debug("Fixing basis in basis aware verification strategies.")
            strategy_pack = strategy_pack.add_basis(basis)
        strategy_pack = strategy_pack.setup_subclass_verification(start_tiling)

        super().__init__(
            start_class=start_tiling,
            strategy_pack=strategy_pack,
            classdb=classdb,
            ruledb=ruledb,
            classqueue=classqueue,
            expand_verified=expand_verified,
            debug=debug,
        )


class LimitedAssumptionTileScope(TileScope):
    """
    A subclass of Tilescope that allows a limit to be set on the maximum number of
    assumptions that appear on any tiling in the universe.
    """

    def __init__(
        self,
        start_class: Union[str, Iterable[Perm], Tiling],
        strategy_pack: TileScopePack,
        max_assumptions: int,
        ignore_full_tiling_assumptions: bool = False,
        **kwargs,
    ) -> None:
        self.max_assumptions = max_assumptions
        super().__init__(
            start_class,
            strategy_pack,
            classdb=TrackedClassDB(),
            **kwargs,
        )
        self.ignore_full_tiling_assumptions = ignore_full_tiling_assumptions

    def _rules_from_strategy(  # type: ignore
        self, comb_class: CombinatorialClassType, strategy: CSSstrategy
    ) -> Iterator[AbstractRule]:
        """
        Yield all the rules given by a strategy/strategy factory whose children all
        satisfy the max_assumptions constraint.
        """
        # pylint: disable=arguments-differ
        def num_child_assumptions(child: Tiling) -> int:
            return sum(
                1
                for ass in child.assumptions
                if (not self.ignore_full_tiling_assumptions)
                or len(ass.gps) != len(child.active_cells)
            )

        for rule in super()._rules_from_strategy(comb_class, strategy):
            if all(
                num_child_assumptions(child) <= self.max_assumptions
                for child in rule.children
            ):
                yield rule


class GuidedSearcher(TileScope):
    def __init__(
        self,
        tilings: Iterable[Tiling],
        basis: Tiling,
        pack: TileScopePack,
        **kwargs,
    ):
        self.tilings = frozenset(t.remove_assumptions() for t in tilings)
        super().__init__(
            basis,
            pack,
            classdb=TrackedClassDB(),
            **kwargs,
        )
        for t in self.tilings:
            class_label = self.classdb.get_label(t)
            is_empty = self.classdb.is_empty(t, class_label)
            if not is_empty:
                self.classqueue.add(class_label)
            self._symmetry_expand(t, class_label)

    def _expand(
        self,
        comb_class: Tiling,
        label: int,
        strategies: Tuple[CSSstrategy, ...],
        inferral: bool,
    ) -> None:
        if comb_class.remove_assumptions() not in self.tilings:
            return
        return super()._expand(comb_class, label, strategies, inferral)

    def _symmetry_expand(self, comb_class: CombinatorialClassType, label: int) -> None:
        sym_labels = set([label])
        for strategy_generator in self.symmetries:
            for start_label, end_labels, rule in self._expand_class_with_strategy(
                comb_class, strategy_generator, label=label
            ):
                sym_label = end_labels[0]
                self.ruledb.add(start_label, (sym_label,), rule)
                self.classqueue.add(sym_label)
                sym_labels.add(sym_label)
        self.symmetry_expanded.update(sym_labels)

    @classmethod
    def from_spec(
        cls, specification: CombinatorialSpecification, pack: TileScopePack
    ) -> "GuidedSearcher":
        tilings = specification.comb_classes()
        root = specification.root
        return cls(tilings, root, pack)

    @classmethod
    def from_uri(cls, URI: str) -> "GuidedSearcher":
        response = requests.get(URI, timeout=10)
        spec = CombinatorialSpecification.from_dict(response.json()["specification"])
        pack = TileScopePack.from_dict(response.json()["pack"]).make_tracked()
        return cls.from_spec(spec, pack)


class TrackedSearcher(LimitedAssumptionTileScope):
    """
    A TileScope that will prioritise expanding tilings whose underlying tilings
    were found at earlier levels. It does this by keeping a queue for each level,
    and adding tilings to the queue that the underlying was first found at.

    The first time a queue changes level, the next level of queue i will be added
    to the curr level of queue i + 1. If `delay_next` is False then it continues
    in this way for future change levels but if it is False (the default) the next
    level of queue i will be added to the curr level of queue i after the first
    change levels.
    """

    def __init__(
        self,
        start_class: Union[str, Iterable[Perm], Tiling],
        strategy_pack: TileScopePack,
        max_assumptions: int,
        delay_next: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(
            start_class,
            strategy_pack,
            max_assumptions=max_assumptions,
            classqueue=TrackedQueue(
                cast(TileScopePack, strategy_pack), self, delay_next
            ),
            **kwargs,
        )


class TrackedDefaultQueue(DefaultQueue):
    def __init__(self, pack: TileScopePack, delay_next: bool):
        super().__init__(pack)
        self.next_curr_level: Optional[Tuple[Deque[int], ...]] = None
        self.delay_next = delay_next

    def is_empty(self) -> bool:
        return bool(
            not self.working and not self.next_level and not any(self.curr_level)
        )

    def set_next_curr_level(self, other: "TrackedDefaultQueue"):
        self.next_curr_level = other.curr_level

    def set_tracked_queue(self, tracked_queue: "TrackedQueue") -> None:
        self._inferral_expanded = tracked_queue.inferral_expanded
        self._initial_expanded = tracked_queue.initial_expanded
        self.ignore = tracked_queue.ignore

    def _change_level(self) -> None:
        assert not self.staging, "Can't change level is staging is not empty"
        assert not self.working, "Can't change level is working is not empty"
        assert not any(self.curr_level), "Can't change level is curr_level is not empty"
        assert self.next_curr_level is not None, "not set the next curr queue"
        if any(self.next_curr_level):
            # this ensures we only change level when the next curr queue is empty
            # and therefore makes sure we never expand a label with the same strategy
            # twice
            raise StopIteration
        self.next_curr_level[0].extend(
            label
            for label, _ in sorted(self.next_level.items(), key=lambda x: -x[1])
            if label not in self.ignore
        )
        self.next_level: CounterType[int] = Counter()
        if not self.delay_next:
            self.next_curr_level = self.curr_level
        if not any(self.curr_level):
            raise StopIteration


class TrackedQueue(CSSQueue):
    def __init__(
        self, pack: TileScopePack, tilescope: TrackedSearcher, delay_next: bool
    ):
        self.tilescope = tilescope
        self.pack = pack
        self.delay_next = delay_next
        self.label_to_underlying: Dict[int, int] = {}
        self._level_first_found: Dict[int, int] = {}
        self._underlyng_labels_per_level: CounterType[int] = Counter()
        self._all_labels_per_level: CounterType[int] = Counter()
        self.inferral_expanded: Set[int] = set()
        self.initial_expanded: Set[int] = set()
        self.ignore: Set[int] = set()
        first_queue = TrackedDefaultQueue(pack, self.delay_next)
        first_queue.set_tracked_queue(self)
        self.queues = [first_queue]
        self.add_new_queue()

        super().__init__(pack)

    @property
    def levels_completed(self):
        """Return the number of levels completed for underlying tilings"""
        return len(self.queues) - 2

    def add_new_queue(self) -> None:
        last_queue = self.queues[-1]
        new_queue = TrackedDefaultQueue(self.pack, self.delay_next)
        new_queue.set_tracked_queue(self)
        last_queue.set_next_curr_level(new_queue)
        self.queues.append(new_queue)

    def get_underlying_label(self, label: int) -> int:
        underlying_label = self.label_to_underlying.get(label)
        if underlying_label is None:
            tiling = self.tilescope.classdb.get_class(label)
            underlying_tiling = tiling.remove_assumptions()
            underlying_label = self.tilescope.classdb.get_label(underlying_tiling)
            self.label_to_underlying[label] = underlying_label
            # count the number of labels that will be added to this level
            self._all_labels_per_level[self.level_first_found(underlying_label)] += 1
        return underlying_label

    def level_first_found(self, label: int) -> int:
        """Return the level that the underlying label was first found at."""
        underlying_label = self.get_underlying_label(label)
        level = self._level_first_found.get(underlying_label)
        if level is None:
            level = len(self.queues) - 2
            self._level_first_found[underlying_label] = level
            # count the number of underlying labels added to this level
            self._underlyng_labels_per_level[level] += 1
        return level

    def add(self, label: int) -> None:
        if label in self.ignore:
            return
        self.queues[self.level_first_found(label)].add(label)

    def set_not_inferrable(self, label: int) -> None:
        self.queues[self.level_first_found(label)].set_not_inferrable(label)

    def set_verified(self, label: int) -> None:
        self.queues[self.level_first_found(label)].set_verified(label)

    def set_stop_yielding(self, label: int) -> None:
        self.queues[self.level_first_found(label)].set_stop_yielding(label)

    def do_level(self) -> Iterator[WorkPacket]:
        raise NotImplementedError

    def status(self) -> str:
        status = f"Queue status (currently on level {self.levels_completed}):\n"
        table: List[Tuple[str, ...]] = []
        working = ("working",) + tuple(
            f"{len(queue.working):,d}" for queue in self.queues[:-1]
        )
        table.append(working)
        for idx in range(len(self.pack.expansion_strats)):
            current = (f"current (set {idx+1})",) + tuple(
                f"{len(queue.curr_level[idx]):,d}" for queue in self.queues[:-1]
            )
            table.append(current)
        nxt = ("next",) + tuple(
            f"{len(queue.next_level):,d}" for queue in self.queues[:-1]
        )
        table.append(nxt)
        status += "    "
        headers = ("Size",) + tuple(
            f"Queue {idx}" for idx in range(len(self.queues) - 1)
        )
        underlying = ("underlying",) + tuple(
            str(self._underlyng_labels_per_level[level])
            for level in range(len(self.queues))
        )
        table.append(underlying)
        all_labels = ("all labels",) + tuple(
            str(self._all_labels_per_level[level]) for level in range(len(self.queues))
        )
        table.append(all_labels)
        table = [headers] + table
        table = list(zip(*table))
        headers = table[0]
        table = table[1:]
        colalign = ("left",) + tuple("right" for _ in headers[1:])
        status += (
            tabulate.tabulate(table, headers=headers, colalign=colalign).replace(
                "\n", "\n    "
            )
            + "\n"
        )
        return status

    def __next__(self) -> WorkPacket:
        for idx, queue in enumerate(self.queues):
            if idx == len(self.queues) - 1:
                if all(queue.is_empty() for queue in self.queues):
                    raise StopIteration
                self.add_new_queue()
            try:
                return next(queue)
            except StopIteration:
                continue
        raise StopIteration("No elements in queue")


class TrackedClassDB(ClassDB[Tiling]):
    def __init__(self) -> None:
        super().__init__(Tiling)
        self.classdb = ClassDB(Tiling)
        self.label_to_tilings: List[bytes] = []
        self.tilings_to_label: Dict[bytes, int] = {}
        self.assumption_type_to_int: Dict[Type[TrackingAssumption], int] = {}
        self.int_to_assumption_type: List[Type[TrackingAssumption]] = []

    def __iter__(self) -> Iterator[int]:
        yield from self.label_to_info

    def __contains__(self, key: Key) -> bool:
        if isinstance(key, Tiling):
            actual_key = self.tiling_to_key(key)
            compressed_key = self._compress_key(actual_key)
            return self.tilings_to_label.get(compressed_key) is not None
        if isinstance(key, int):
            return 0 <= key < len(self.label_to_tilings)
        raise ValueError("Invalid key")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TrackedClassDB):
            return NotImplemented
        return bool(
            self.classdb == other.classdb
            and self.label_to_tilings == other.label_to_tilings
            and self.tilings_to_label == other.tilings_to_label
        )

    def tiling_to_key(self, tiling: Tiling) -> TrackedClassDBKey:
        """
        Converts a tiling to its corresponding key.
        """
        underlying_label = self.classdb.get_label(tiling.remove_assumptions())
        assumption_keys = tuple(
            self.assumption_to_key(ass) for ass in tiling.assumptions
        )
        return (underlying_label, assumption_keys)

    def assumption_to_key(self, ass: TrackingAssumption) -> TrackedClassAssumption:
        """
        Determines the type of the assumption and retrieves the int representing
        that type from the appropriate class variables, and then apprends the cells.
        """
        try:
            ass_type_int = self.assumption_type_to_int[type(ass)]
        except KeyError:
            ass_type_int = len(self.int_to_assumption_type)
            assert ass_type_int < 256
            self.int_to_assumption_type.append(type(ass))
            self.assumption_type_to_int[type(ass)] = ass_type_int
        return (ass_type_int, ass.get_cells())

    def key_to_tiling(self, key: TrackedClassDBKey) -> Tiling:
        """
        Converts a key back to a Tiling.
        """
        return self.classdb.get_class(key[0]).add_assumptions(
            (
                self.int_to_assumption_type[ass_key[0]].from_cells(ass_key[1])
                for ass_key in key[1]
            ),
            clean=False,
        )

    @staticmethod
    def _compress_key(key: TrackedClassDBKey) -> bytes:
        # Assumes there are fewer than 256 assumptions
        # Assumes every assumption covers fewer than 256 cells
        # Assumes the positions in an assumption have value < 256

        def int_to_bytes(n: int) -> List[int]:
            """
            Converts an int to a list of ints all in [0 .. 255] ready for
            byte compression. First entry is the number of bytes needed (assumes < 256),
            remaining entries the bytes composing the int from lowest byte up to largest
            byte.
            """
            bytes_needed = max(math.ceil(n.bit_length() / 8), 1)
            result: List[int] = [bytes_needed]
            while n >= 2**8:
                result.append(n & 0xFF)
                n = n >> 8
            result.append(n)
            return result

        def _compress_assumption(ass_key: TrackedClassAssumption) -> List[int]:
            type_int, cells = ass_key
            assert type_int < 256
            assert len(cells) < 256
            assert all(cell[0] < 256 and cell[1] < 256 for cell in cells)

            result = [type_int]
            result.append(len(cells))
            result.extend(itertools.chain(*cells))
            return result

        result: List[int] = int_to_bytes(key[0])
        result.extend(
            itertools.chain.from_iterable(
                _compress_assumption(ass_key) for ass_key in key[1]
            )
        )
        compressed_key = array("B", result).tobytes()
        return compressed_key

    @staticmethod
    def _decompress_key(compressed_key: bytes) -> TrackedClassDBKey:
        def int_from_bytes(n: array) -> int:
            """
            Converts a list of ints to a single int assuming the first entry is the
            lowest byte and so on.
            """
            result = n[0]
            for idx in range(1, len(n)):
                result |= n[idx] << (8 * idx)
            return cast(int, result)

        def _decompress_tuple_of_cells(
            compressed_cells: array,
        ) -> Tuple[Cell, ...]:
            """
            compressed_cells is a list of 2*i bytes, each of which is a coordinates
            """
            vals = iter(compressed_cells)
            return tuple(
                (next(vals), next(vals)) for _ in range(len(compressed_cells) // 2)
            )

        vals = array("B", compressed_key)
        offset = 0

        num_bytes_int = vals[offset]
        offset += 1
        label = int_from_bytes(vals[offset : offset + num_bytes_int])
        offset += num_bytes_int

        tuples_of_cells = []
        while offset < len(vals):
            type_int, num_cells = vals[offset : offset + 2]
            offset += 2
            tuples_of_cells.append(
                (
                    type_int,
                    _decompress_tuple_of_cells(vals[offset : offset + 2 * num_cells]),
                )
            )
            offset += 2 * num_cells

        return (label, tuple(tuples_of_cells))

    def add(self, comb_class: ClassKey, compressed: bool = False) -> None:
        """
        Adds a Tiling to the classdb
        """
        if compressed:
            raise NotImplementedError
        if isinstance(comb_class, Tiling):
            key = self.tiling_to_key(comb_class)
            compressed_key = self._compress_key(key)
            if compressed_key not in self.tilings_to_label:
                self.label_to_tilings.append(compressed_key)
                self.tilings_to_label[compressed_key] = len(self.tilings_to_label)

    def _get_info(self, key: Key) -> Info:
        """
        Return the "Info" object corresponding to the key, which is
        either a Tiling or an integer
        """
        # pylint: disable=protected-access
        if isinstance(key, Tiling):
            actual_key = self.tiling_to_key(key)
            compressed_key = self._compress_key(actual_key)
            if compressed_key not in self.tilings_to_label:
                self.add(key)
            info: Optional[Info] = self.classdb._get_info(actual_key[0])
            if info is None:
                raise ValueError("Invalid key")
            info = Info(
                key,
                self.tilings_to_label[compressed_key],
                info.empty,
            )
        elif isinstance(key, int):
            if not 0 <= key < len(self.label_to_tilings):
                raise KeyError("Key not in ClassDB")
            tiling_key = self._decompress_key(self.label_to_tilings[key])
            info = self.classdb.label_to_info.get(tiling_key[0])
            if info is None:
                raise ValueError("Invalid key")
            info = Info(
                self.key_to_tiling(tiling_key),
                key,
                info.empty,
            )
        else:
            raise TypeError()
        return info

    def get_class(self, key: Key) -> Tiling:
        """
        Return combinatorial class of key.
        """
        info = self._get_info(key)
        return cast(Tiling, info.comb_class)

    def is_empty(self, comb_class: Tiling, label: Optional[int] = None) -> bool:
        """
        Return True if combinatorial class is set to be empty, False if not.
        """
        return bool(self.classdb.is_empty(comb_class.remove_assumptions()))

    def set_empty(self, key: Key, empty: bool = True) -> None:
        """
        Set a class to be empty.
        """
        if isinstance(key, int):
            if 0 <= key < len(self.label_to_tilings):
                underlying_label, _ = self._decompress_key(self.label_to_tilings[key])
            else:
                raise ValueError("Invalid key")
        elif isinstance(key, Tiling):
            underlying_label = self.classdb.get_label(key.remove_assumptions())
        else:
            raise ValueError("Invalid key")
        self.classdb.set_empty(underlying_label, empty)

    def status(self) -> str:
        """
        Return a string with the current status of the run.
        """
        status = self.classdb.status()
        status = status.replace("combinatorial classes", "underlying tilings")
        tilings = "\n\tTotal number of tilings found is"
        tilings += f" {len(self.label_to_tilings):,d}"
        status = status.replace("ClassDB status:", "TrackedClassDB status:" + tilings)
        return status + "\n"



from itertools import chain
from typing import Dict, Iterator, List, Optional, Tuple

from comb_spec_searcher import DisjointUnionStrategy, StrategyFactory
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.strategies import Rule
from comb_spec_searcher.strategies.rule import EquivalencePathRule
from permuta import Perm
from tilings import GriddedPerm, Tiling
from tilings.algorithms import Fusion

from .symmetry import TilingRotate90, TilingRotate270


class GeneralizedSlidingStrategy(DisjointUnionStrategy[Tiling, GriddedPerm]):
    """
    A strategy that slides column idx and idx + 1.
    """

    def __init__(self, idx: int, rotate: bool = False):
        super().__init__()
        self.idx = idx
        self.rotate = rotate

    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling]:
        if self.can_slide(comb_class):
            return (self.slide_tiling(comb_class),)
        raise StrategyDoesNotApply(f"Sliding idx {self.idx} does not apply")

    def can_slide(self, comb_class: Tiling) -> bool:
        return MonotoneSlidingFactory.can_slide_col(
            self.idx, comb_class.rotate270() if self.rotate else comb_class
        )

    def slide_tiling(self, comb_class: Tiling) -> Tiling:
        if self.rotate:
            comb_class = comb_class.rotate270()
        child = Tiling(
            self.slide_gps(comb_class.obstructions),
            map(self.slide_gps, comb_class.requirements),
            [ass.__class__(self.slide_gps(ass.gps)) for ass in comb_class.assumptions],
        )
        if self.rotate:
            child = child.rotate90()
        return child

    def formal_step(self) -> str:
        return f"Sliding index {self.idx}"

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        raise NotImplementedError

    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm], Optional[GriddedPerm]]:
        raise NotImplementedError

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str], ...]:
        if not comb_class.extra_parameters:
            return super().extra_parameters(comb_class, children)
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        if self.rotate:
            rules: List[Rule[Tiling, GriddedPerm]] = []
            parent = comb_class
            for strategy in [
                TilingRotate270(),
                GeneralizedSlidingStrategy(self.idx, False),
                TilingRotate90(),
            ]:
                rules.append(strategy(parent))
                parent = rules[-1].children[0]
            return EquivalencePathRule(rules).constructor.extra_parameters
        child = children[0]
        return (
            {
                comb_class.get_assumption_parameter(
                    ass
                ): child.get_assumption_parameter(
                    ass.__class__(self.slide_gps(ass.gps))
                )
                for ass in comb_class.assumptions
            },
        )

    def slide_gp(self, gp: GriddedPerm) -> GriddedPerm:
        pos = sorted(
            x if x < self.idx or x > self.idx + 1 else x + 1 if x == self.idx else x - 1
            for x, _ in gp.pos
        )
        return GriddedPerm(gp.patt, ((x, 0) for x in pos))

    def slide_gps(self, gps: Tuple[GriddedPerm, ...]) -> Tuple[GriddedPerm, ...]:
        return tuple(map(self.slide_gp, gps))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(idx={self.idx}, rotate={self.rotate})"

    def __str__(self) -> str:
        return f"slide column {self.idx} and {self.idx + 1}"

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d["idx"] = self.idx
        d["rotate"] = self.rotate
        return d

    @classmethod
    def from_dict(cls, d: dict):
        return cls(d["idx"], d["rotate"])


class MonotoneSlidingFactory(StrategyFactory[Tiling]):
    """
    A factory that creates rules that swaps neighbouring cells if they
    are 'monotone' fusable, i.e., they are a generalized fusion with
    a monotone local extra obstruction.

    This is only looks at n x 1 and 1 x n tilings.
    """

    def __call__(self, comb_class: Tiling) -> Iterator[Rule]:
        parent = comb_class
        rotate = False
        if (
            not comb_class.dimensions[1] == 1
            and comb_class.dimensions[0] == 1
            and not comb_class.requirements
        ):
            comb_class = comb_class.rotate270()
            rotate = True

        if comb_class.dimensions[1] == 1 and not comb_class.requirements:
            # TODO: allow requirements outside of sliding region
            for col in range(comb_class.dimensions[0] - 1):
                if self.can_slide_col(col, comb_class):
                    strategy = GeneralizedSlidingStrategy(col, rotate)
                    child = strategy.slide_tiling(parent)
                    yield strategy(parent, (child,))

    @staticmethod
    def can_slide_col(col: int, comb_class: Tiling) -> bool:
        """
        Return True if the column can be slid.
        """
        local_cells = (
            comb_class.cell_basis()[(col, 0)][0],
            comb_class.cell_basis()[(col + 1, 0)][0],
        )
        if MonotoneSlidingFactory.valid_monotone_sliding_region(
            col, local_cells, comb_class
        ):
            # Check the fusability condition
            shortest = (
                col if len(local_cells[0][0]) <= len(local_cells[1][0]) else col + 1
            )
            algo = Fusion(comb_class, col_idx=col)
            fused_obs = tuple(
                algo.fuse_gridded_perm(gp)
                for gp in comb_class.obstructions
                if not all(x == shortest for x, _ in gp.pos)
            )
            unfused_obs = tuple(
                chain.from_iterable(algo.unfuse_gridded_perm(gp) for gp in fused_obs)
            )
            return comb_class == comb_class.add_obstructions(unfused_obs)
        return False

    @staticmethod
    def valid_monotone_sliding_region(
        col: int, local_cells: Tuple[List[Perm], List[Perm]], comb_class: Tiling
    ) -> bool:
        """
        Return True if the region is a possible valid monotone sliding region.

        That is:
            - neighbouring cells are both increasing or decreasing.
            - the values of non-local obstructions in sliding region are
            monotone and consecutive in value.
        """

        def consecutive_value(col: int, tiling: Tiling, incr: bool = True) -> bool:
            """
            Return True if the values in the column are consecutive,
            and increasing or decreasing if incr is True or False.
            """
            for gp in tiling.obstructions:
                if any(x not in (col, col + 1) for x, _ in gp.pos):
                    points = chain(gp.get_points_col(col), gp.get_points_col(col + 1))
                    values = [y for _, y in points]
                    if incr and not all(x + 1 == y for x, y in zip(values, values[1:])):
                        return False
                    if not incr and not all(
                        x - 1 == y for x, y in zip(values, values[1:])
                    ):
                        return False
            return True

        return (len(local_cells[0]) == 1 and len(local_cells[1]) == 1) and (
            (  # both cells are increasing, and consecutive values are increasing
                local_cells[0][0].is_increasing()
                and local_cells[1][0].is_increasing()
                and consecutive_value(col, comb_class)
            )
            or (  # both cells are decreasing, and consecutive values are decreasing
                (
                    local_cells[0][0].is_decreasing()
                    and local_cells[1][0].is_decreasing()
                    and consecutive_value(col, comb_class, False)
                )
            )
        )

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def __str__(self):
        return "monotone sliding"

    @classmethod
    def from_dict(cls, d: dict):
        return cls()



from collections import defaultdict
from itertools import product
from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Set, Tuple

from permuta import Perm
from tilings import GriddedPerm

if TYPE_CHECKING:
    from tilings import Tiling

Cell = Tuple[int, int]


class ObstructionTransitivity:
    """
    The obstruction transitivity algorithm.

    The obstruction transitivity considers all length 2 obstructions with both
    points in the same row or same column. By considering these length 2
    obstructions in similar manner as the row and column separation, as
    inequality relations. When the obstructions use a positive cell,
    transitivity applies, i.e. if a < b < c and b is positive, then a < c.
    """

    def __init__(self, tiling: "Tiling") -> None:
        self._tiling = tiling
        self._colineq: Optional[Dict[int, Set[Tuple[int, int]]]] = None
        self._rowineq: Optional[Dict[int, Set[Tuple[int, int]]]] = None
        self._positive_cells_col: Optional[Dict[int, List[int]]] = None
        self._positive_cells_row: Optional[Dict[int, List[int]]] = None
        self._new_ineq: Optional[List[Tuple[Tuple[int, int], Tuple[int, int]]]] = None

    def positive_cells_col(self, col_index: int) -> List[int]:
        """
        Return the row index of all positive cells in the columns.

        OUTPUT:

            List of integers.
        """
        if self._positive_cells_col is None:
            self._init_positive_cells()
        assert self._positive_cells_col is not None
        return self._positive_cells_col[col_index]

    def positive_cells_row(self, row_index: int) -> List[int]:
        """
        Return the column index of all positive cells in the row.

        OUTPUT:

            List of integers.
        Return a dictionary of positive cells for each column.
        """
        if self._positive_cells_row is None:
            self._init_positive_cells()
        assert self._positive_cells_row is not None
        return self._positive_cells_row[row_index]

    def ineq_col(self, col_index: int) -> Set[Tuple[int, int]]:
        """
        Returns a set of inequalities in the specified column.

        An inequality is a tuple of integer (r1, r2) such that r1 < r2.
        """
        if self._colineq is None:
            self._init_ineq()
        assert self._colineq is not None
        return self._colineq[col_index]

    def ineq_row(self, row_index: int) -> Set[Tuple[int, int]]:
        """
        Returns a set of inequalities in the specified row.

        An inequality is a tuple of integer (c1, c2) such that c1 < c2.
        """
        if self._rowineq is None:
            self._init_ineq()
        assert self._rowineq is not None
        return self._rowineq[row_index]

    def _init_positive_cells(self) -> None:
        """
        Fill the dictionary of positive cells by rows and columns.
        """
        self._positive_cells_row = defaultdict(list)
        self._positive_cells_col = defaultdict(list)
        for cell in self._tiling.positive_cells:
            self._positive_cells_col[cell[0]].append(cell[1])
            self._positive_cells_row[cell[1]].append(cell[0])

    def _init_ineq(self) -> None:
        """
        Fill the dictionary of row and column inequalities.
        """
        self._colineq = defaultdict(set)
        self._rowineq = defaultdict(set)
        len2_obs = (ob for ob in self._tiling.obstructions if len(ob) == 2)
        ineq_obs = (ob for ob in len2_obs if not ob.is_localized())
        for ob in ineq_obs:
            leftcol, rightcol = ob.pos[0][0], ob.pos[1][0]
            leftrow, rightrow = ob.pos[0][1], ob.pos[1][1]
            # In same column
            if leftcol == rightcol:
                self._colineq[leftcol].add((rightrow, leftrow))
            # In same row
            elif ob.pos[0][1] == ob.pos[1][1]:
                if ob.patt == Perm((0, 1)):
                    self._rowineq[leftrow].add((rightcol, leftcol))
                else:
                    self._rowineq[leftrow].add((leftcol, rightcol))

    @staticmethod
    def ineq_ob(ineq) -> GriddedPerm:
        """
        Given an inequality of cells compute an obstruction.

        The inequalities are given as tuple `(cell1, cell2)` where `cell1` is
        strictly smaller than `cell2`.
        """
        left, right = ineq
        if left == right:
            return GriddedPerm((0,), (left,))
        if left[0] == right[0]:
            # same column
            if left[1] < right[1]:
                return GriddedPerm((1, 0), [right, left])
            return GriddedPerm((0, 1), [right, left])
        if left[1] == right[1]:
            # same row
            if left[0] < right[0]:
                return GriddedPerm((1, 0), [left, right])
            return GriddedPerm((0, 1), [right, left])
        raise ValueError(
            f"Can not construct an obstruction from inequality {left} < {right}"
        )

    @staticmethod
    def ineq_closure(positive_cells: Iterable[int], ineqs: Set[Cell]) -> Set[Cell]:
        """
        Computes the transitive closure over positive cells.

        Given a list of inequalities and positive cells, a new set of
        inequalities is computed. For every positive cell c, when g < c < l the
        relation g < l is added if not already there.

        The list of new inequalities is returned.
        """
        gtlist: Dict[int, List[int]] = defaultdict(list)
        ltlist: Dict[int, List[int]] = defaultdict(list)
        for left, right in ineqs:
            ltlist[left].append(right)
            gtlist[right].append(left)
        to_analyse = set(positive_cells)
        ineqs = set(ineqs)
        newineqs = set()
        while to_analyse:
            cur = to_analyse.pop()
            not_existing_ineq = filter(
                lambda x: x not in ineqs, product(gtlist[cur], ltlist[cur])
            )
            for gt, lt in not_existing_ineq:
                gtlist[lt].append(gt)
                ltlist[gt].append(lt)
                ineqs.add((gt, lt))
                newineqs.add((gt, lt))
                if lt in positive_cells:
                    to_analyse.add(lt)
                if gt in positive_cells:
                    to_analyse.add(gt)
        return newineqs

    def new_ineq(
        self,
    ) -> List[Tuple[Tuple[int, int], Tuple[int, int]]]:
        """
        Compute the new inequalities.
        """
        if self._new_ineq is not None:
            return self._new_ineq
        newineqs: List[Tuple[Tuple[int, int], Tuple[int, int]]] = []
        ncol, nrow = self._tiling.dimensions
        for col in range(ncol):
            ineqs = self.ineq_col(col)
            pos_cells = self.positive_cells_col(col)
            for left, right in self.ineq_closure(pos_cells, ineqs):
                newineqs.append(((col, left), (col, right)))
        for row in range(nrow):
            ineqs = self.ineq_row(row)
            pos_cells = self.positive_cells_row(row)
            for left, right in self.ineq_closure(pos_cells, ineqs):
                newineqs.append(((left, row), (right, row)))
        self._new_ineq = newineqs
        return self._new_ineq

    def new_obs(self) -> Tuple[GriddedPerm, ...]:
        """Return the obstructions that are implied by the method."""
        return tuple(map(self.ineq_ob, self.new_ineq()))

    def obstruction_transitivity(self) -> "Tiling":
        """
        Return the tiling with the new obstructions.
        """
        return self._tiling.add_obstructions(self.new_obs())

    def __str__(self) -> str:
        s = "ObstructionTransitivity object for the tiling:\n"
        s += self._tiling.__str__()
        return s
